"""The module containing the base class for data converter."""
from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path

import vaex

from mleko.cache.lru_cache_mixin import LRUCacheMixin
from mleko.dataset.data_schema import DataSchema


class BaseConverter(LRUCacheMixin, ABC):
    """Abstract base class for data converter."""

    def __init__(self, cache_directory: str | Path, cache_size: int):
        """Initialize the `BaseConverter`.

        The `cache_size` is the maximum number of cache entries, and the cache will be cleared if the number of
        entries exceeds this value.

        Args:
            cache_directory: The directory to store the cache in.
            cache_size: The maximum number of cache entries.
        """
        LRUCacheMixin.__init__(self, cache_directory, cache_size)

    @abstractmethod
    def convert(
        self, file_paths: list[Path] | list[str], cache_group: str | None = None, force_recompute: bool = False
    ) -> tuple[DataSchema, vaex.DataFrame]:
        """Abstract method to convert the input file paths to the desired output format.

        Args:
            file_paths: A list of input file paths to be converted.
            cache_group: The cache group to use.
            force_recompute: If set to True, forces recomputation and ignores the cache.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError



"""Data management module for handling data used in the pipeline.

This module provides the DataContainer class, which serves as a common interface for various types of data,
enforcing shared structure and behavior. The goal is to facilitate data handling throughout the pipeline.
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


@dataclass
class DataContainer:
    """Class for holding data used in the pipeline.

    This class serves as a common interface and can be extended to enforce
    a shared structure or behavior across different types of data.

    Examples:
        >>> data_container = DataContainer()
        >>> data_container.data["my_data"] = [Path("path/to/data")]
        >>> data_container.data["my_data"]
        [Path("path/to/data")]
    """

    data: dict[str, Any] = field(default_factory=dict)
    """The data stored in the DataContainer.

    The data is stored in a dictionary, where the keys are the names of the data and the values are the data itself.
    """

    def __repr__(self) -> str:
        """Get string representation of DataContainer.

        Returns:
            String representation of the DataContainer, including data type and stored data.
        """
        data_type = type(self.data).__name__
        return f"<DataContainer: data_type={data_type}, data={self.data}>"



"""Module for the max-abs scaler transformer."""
from __future__ import annotations

from pathlib import Path
from typing import Hashable

import vaex
import vaex.ml

from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_transformer import BaseTransformer


logger = CustomLogger()
"""A module-level logger for the module."""


class MaxAbsScalerTransformer(BaseTransformer):
    """Transforms features using maximum absolute scaling."""

    @auto_repr
    def __init__(
        self,
        cache_directory: str | Path,
        features: list[str] | tuple[str, ...],
        cache_size: int = 1,
    ) -> None:
        """Initializes the max absolute scaler transformer.

        The max absolute scaler transformer will scale each feature by its maximum absolute value. This transformer
        will not shift/center the data, and thus will not destroy any sparsity.

        Warning:
            Should only be used with numerical features. There should be no missing values in the features
            or an error will be raised.

        Args:
            cache_directory: Directory where the cache will be stored locally.
            features: List of feature names to be used by the transformer.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.dataset.transform import MaxAbsScalerTransformer
            >>> df = vaex.from_arrays(
            ...     a=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            ...     b=[-1, -2, -3, -4, -5, 0, 1, 2, 3, 4]
            ... )
            >>> ds = DataSchema(
            ...     numerical=["a", "b"],
            ... )
            >>> _, _, df = MaxAbsScalerTransformer(
            ...     cache_directory=".",
            ...     features=["a", "b"],
            ... ).fit_transform(ds, df)
            >>> df["a"].tolist()
            [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
            >>> df["b"].tolist()
            [-0.2, -0.4, -0.6, -0.8, -1.0, 0.0, 0.2, 0.4, 0.6, 0.8]
        """
        super().__init__(cache_directory, features, cache_size)
        self._transformer = vaex.ml.MaxAbsScaler(features=self._features, prefix="")

    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.ml.MaxAbsScaler]:
        """Fits the transformer on the given DataFrame.

        Args:
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to fit the transformer on.

        Returns:
            Updated data schema and fitted transformer.
        """
        logger.info(f"Fitting max-abs scaler transformer ({len(self._features)}): {self._features}.")
        self._transformer.fit(dataframe)

        return data_schema, self._transformer

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Transforms the features in the DataFrame using max-abs scaling.

        Args:
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to transform.

        Returns:
            Updated data schema and transformed DataFrame.
        """
        logger.info(f"Transforming features using max-abs scaling ({len(self._features)}): {self._features}.")
        transformed_df = self._transformer.transform(dataframe)

        return data_schema, transformed_df

    def _fingerprint(self) -> Hashable:
        """Returns the fingerprint of the transformer.

        Returns:
            A hashable object that uniquely identifies the transformer.
        """
        return super()._fingerprint()



"""Provides a flexible and customizable Pipeline class for managing and executing a series of data processing steps.

The module defines a Pipeline class that allows for the creation of a sequence of data processing steps. It is
designed for building complex data processing workflows by chaining together custom `PipelineStep` instances.
The Pipeline class encapsulates the ordered sequence of steps added to it and provides methods to manage, execute,
and visualize these steps. Each step's output is passed as input to the next step, effectively managing the flow
of data through the processing pipeline.
"""
from __future__ import annotations

from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.custom_logger import CustomLogger


logger = CustomLogger()
"""A CustomLogger instance that's used throughout the module for logging."""


class Pipeline:
    """Encapsulates a pipeline that manages and executes a series of data processing steps in a defined order."""

    def __init__(self, steps: list[PipelineStep] | None = None) -> None:
        """Creates a new Pipeline instance, initializing it with a list of steps or an empty list.

        Note:
            The steps list can be provided as an argument to the constructor, or the pipeline can be initialized
            with an empty list of steps and have them added later using the `add_step` method. This allows for
            more flexibility in the creation of the pipeline, as steps can be added dynamically.

        Args:
            steps: An optional list of `PipelineStep` instances that define the data processing steps in the
                   pipeline. If not provided, the pipeline will be initialized with an empty list of steps,
                   allowing steps to be added later using the `add_step` method.
        """
        self._steps = steps if steps is not None else []

    def __repr__(self) -> str:
        """Returns a string representation of the Pipeline, including the ordered list of steps.

        Returns:
            String representaition of Pipeline that includes the class name and each step in the order they appear
            in the pipeline, numbered for easier identification.
        """
        cls_name = type(self).__name__
        steps_str = "\n".join([f"  {index + 1}. {step!r}" for index, step in enumerate(self._steps)])
        return f"{cls_name}:\n{steps_str}"

    def add_step(self, step: PipelineStep) -> None:
        """Appends a new PipelineStep to the end of the pipeline, extending the processing sequence.

        Adding a step to the pipeline implies that it will be executed after all the steps previously
        appended to the pipeline when calling the `run` method.

        Args:
            step: The PipelineStep instance to be added at the end of the pipeline's steps list.
        """
        self._steps.append(step)

    def run(self, data_container: DataContainer | None = None, force_recompute: bool = False) -> DataContainer:
        """Executes the pipeline steps in the order they were added, passing output from one to the next.

        Processes the initial given data or an empty data container through each step in the pipeline.
        The output of each step is passed as input to the next step, allowing the given input to be transformed
        through the whole sequence of steps.

        Args:
            data_container: An optional DataContainer instance carrying the input data to be processed by the
                            first step in the pipeline. If not provided, an empty DataContainer instance will be
                            created automatically, and the first step's execute method must handle it.
            force_recompute: Whether to force the pipeline to recompute its output, even if it already exists.

        Returns:
            The output as a DataContainer instance from the last step in the pipeline after processing the data.
        """
        if data_container is None:
            logger.info("No data container provided. Creating an empty one.")
            data_container = DataContainer()

        for i, step in enumerate(self._steps):
            logger.info(f"Executing step {i+1}/{len(self._steps)}: {step.__class__.__name__}.")
            data_container = step.execute(data_container, force_recompute)
            logger.info(f"Finished step {i+1}/{len(self._steps)} execution.")
        return data_container



"""This module defines the abstract base class for pipeline steps in a data processing pipeline.

The module provides a standard interface for implementing data processing steps as part of a larger pipeline,
via the `PipelineStep` abstract base class. Each `PipelineStep` subclass should have a specific purpose and
should be able to run independently or as part of the pipeline.
"""
from __future__ import annotations

from abc import ABC, abstractmethod

from mleko.pipeline.data_container import DataContainer


class PipelineStep(ABC):
    """Base class for all pipeline steps, ensuring the standardized interface for performing data processing operations.

    Descendants of this class must implement the `execute` method, which carries out the data processing operation
    related to the step.

    Note:
        The _num_inputs and _num_outputs attributes are used to validate the number of inputs and outputs, respectively,
        for each step. These attributes are set by the subclasses, and should not be modified by the user.
        When implementing a new step, you should set these attributes to the expected number of inputs and outputs.
    """

    _num_inputs: int
    """Number of inputs expected by the PipelineStep."""

    _num_outputs: int
    """Number of outputs expected by the PipelineStep."""

    def __init__(
        self,
        inputs: list[str] | tuple[str, ...] | tuple[()],
        outputs: list[str] | tuple[str, ...] | tuple[()],
        cache_group: str | None,
    ) -> None:
        """Initialize a new PipelineStep with the provided input and output keys.

        Args:
            inputs: List or tuple of input keys expected by this step.
            outputs: List or tuple of output keys produced by this step.
            cache_group: The cache group to use.
        """
        self._inputs = tuple(inputs)
        self._outputs = tuple(outputs)
        self._cache_group = cache_group

        self._validate_inputs()
        self._validate_outputs()

    @abstractmethod
    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Execute the data processing operation associated with this pipeline step.

        The `execute` method is the main entry point for the data processing operation associated with this step.
        It receives a `DataContainer` instance as input, containing the data to be processed by this step.
        The method should perform the processing operation and return the processed data as a `DataContainer` instance.

        Args:
            data_container: Input data for this step's processing operation.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            NotImplementedError: Must be implemented by subclass.
        """
        raise NotImplementedError

    def _validate_inputs(self) -> None:
        """Check the input keys for compliance with this step's requirements.

        Raises:
            ValueError: If the PipelineStep has an invalid number of inputs.
        """
        if len(self._inputs) != self._num_inputs:
            raise ValueError(f"{self.__class__.__name__} must have exactly {self._num_inputs} input(s).")

    def _validate_outputs(self) -> None:
        """Check the output keys for compliance with this step's requirements.

        Raises:
            ValueError: If the PipelineStep has an invalid number of outputs.
        """
        if len(self._outputs) != self._num_outputs:
            raise ValueError(f"{self.__class__.__name__} must have exactly {self._num_outputs} output(s).")



"""This module provides helper functions for `tqdm` progress bars."""
from __future__ import annotations

from typing import Callable

from tqdm.auto import tqdm


def set_tqdm_percent_wrapper(pbar: tqdm) -> Callable[[float], None]:
    """Return a function to set the percentage of a `tqdm` progress bar instead of incrementing it.

    This function returns a function that can be used to set the percentage of a `tqdm` progress bar instead of
    incrementing it. This is useful when the progress bar is used to track the progress of a task that is not
    necessarily linear, such as a hyperparameter search.

    Args:
        pbar: A `tqdm` progress bar instance.

    Returns:
        A function that sets the percentage based on the float value passed as a parameter.

    Example:
        >>> from tqdm.auto import tqdm
        >>> from mleko.utils import set_tqdm_percent_wrapper
        >>> pbar = tqdm(total=100)
        >>> set_percent = set_tqdm_percent_wrapper(pbar)
        >>> set_percent(0.5)
        >>> pbar.n
        50
    """

    def set_percent(fraction: float) -> None:
        """Set the percentage of the progress bar based on the provided fraction.

        Args:
            fraction: A float value in the range [0, 1] representing the completion percentage of the task.
        """
        pbar.n = int(max(0, min(fraction, 1)) * 100)
        pbar.refresh()

    return set_percent



"""Module containing the FeatureSelectStep class.""" ""
from __future__ import annotations

from typing import Literal

from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.dataset.feature_select.base_feature_selector import BaseFeatureSelector
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class FeatureSelectStep(PipelineStep):
    """Pipeline step that selects a subset of features from a DataFrame."""

    _num_inputs = 2
    """Number of inputs expected by the FeatureSelectStep."""

    _num_outputs = 2
    """Number of outputs expected by the FeatureSelectStep."""

    @auto_repr
    def __init__(
        self,
        feature_selector: BaseFeatureSelector,
        action: Literal["fit", "transform", "fit_transform"],
        inputs: list[str] | tuple[str, ...] | tuple[()] = (),
        outputs: list[str] | tuple[str, ...] | tuple[()] = (),
        cache_group: str | None = None,
    ) -> None:
        """Initialize the FeatureSelectStep with the specified feature selector.

        Args:
            feature_selector: The FeatureSelector responsible for handling feature selection.
            action: The action to perform, one of "fit", "transform", or "fit_transform".
            inputs: List or tuple of input keys expected by this step. Should contain two keys,
                corresponding to the DataSchema and DataFrame to be transformed.
            outputs: List or tuple of output keys produced by this step. If the action is "fit" or "transform",
                should contain two keys, corresponding to the DataSchema and DataFrame after transformation. If the
                action is "fit_transform", should contain three keys, corresponding to the DataSchema, the
                FeatureSelector, and the DataFrame after transformation.
            cache_group: The cache group to use.

        Raises:
            ValueError: If action is not one of "fit", "transform", or "fit_transform".
        """
        if action not in ("fit", "transform", "fit_transform"):
            raise ValueError(
                f"Invalid action: {action}. Expected one of 'fit', 'transform', or 'fit_transform'."
            )  # pragma: no cover

        if action == "fit_transform":
            self._num_outputs = 3

        super().__init__(inputs, outputs, cache_group)
        self._feature_selector = feature_selector
        self._action = action

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform feature selection using the configured feature selector.

        Args:
            data_container: Contains the input DataFrame.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            ValueError: If the data container does not contain the expected inputs.

        Returns:
            A DataContainer containing the result depending on the action.
        """
        data_schema = data_container.data[self._inputs[0]]
        if not isinstance(data_schema, DataSchema):
            raise ValueError(f"Invalid data type: {type(data_schema)}. Expected DataSchema.")

        dataframe = data_container.data[self._inputs[1]]
        if not isinstance(dataframe, DataFrame):
            raise ValueError(f"Invalid data type: {type(dataframe)}. Expected vaex DataFrame.")

        if self._action == "fit":
            ds, feature_selector = self._feature_selector.fit(
                data_schema, dataframe, self._cache_group, force_recompute
            )
            data_container.data[self._outputs[0]] = ds
            data_container.data[self._outputs[1]] = feature_selector
        elif self._action == "transform":
            ds, df = self._feature_selector.transform(data_schema, dataframe, self._cache_group, force_recompute)
            data_container.data[self._outputs[0]] = ds
            data_container.data[self._outputs[1]] = df
        elif self._action == "fit_transform":
            ds, feature_selector, df = self._feature_selector.fit_transform(
                data_schema, dataframe, self._cache_group, force_recompute
            )
            data_container.data[self._outputs[0]] = ds
            data_container.data[self._outputs[1]] = feature_selector
            data_container.data[self._outputs[2]] = df

        return data_container



"""`mleko`: Streamlining Machine Learning Pipelines in Python.

Simplify and accelerate your machine learning development with `mleko`. Designed with modularity and
customization in mind, it seamlessly integrates into your existing workflows. Its robust caching system
optimizes performance, taking you from data ingestion to finalized models with unparalleled efficiency.

`mleko` is engineered to address the end-to-end needs of machine learning pipelines, providing robust,
scalable solutions for data science challenges:
* Ingest: Seamlessly integrates with data sources like AWS S3 and Kaggle, offering hassle-free data
    ingestion and compatibility.
* Convert: Specializes in data format transformations, prominently featuring high-performance
    conversions from CSV to Vaex DataFrame, to make your data pipeline-ready.
* Split: Employs sophisticated data partitioning algorithms, allowing you to segment DataFrames into
    train, test, and validation sets for effective model training and evaluation.
* Feature Selection: Equipped with a suite of feature selection techniques, `mleko` enables model
    performance by focusing on the most impactful variables.
* Transformation: Facilitates data manipulations such as Frequency Encoding and Standardization, ensuring
    that your data conforms to the prerequisites of the machine learning algorithms.
* Model: Provides a core set of functionalities for machine learning models, including in-built support
    for hyperparameter tuning, thereby streamlining the path from data to deployable model.
* Pipeline: Unifies the entire workflow into an intuitive directed acyclic graph (DAG) architecture,
    promoting reproducibility and reducing iteration time and time-to-market for machine learning models.

By integrating these features, `mleko` serves as a comprehensive toolkit for machine learning
practitioners looking to build robust models efficiently.
"""
from __future__ import annotations


__version__ = "1.2.5"



"""The subpackage provides functionality for splitting `vaex` DataFrames into two separate DataFrames.

The subpackage contains the following splitter classes:
    - `BaseSplitter`: The abstract base class for all splitters.
    - `RandomSplitter`: A splitter for splitting `vaex` DataFrames into two random parts.
    - `ExpressionSplitter`: A splitter for splitting `vaex` DataFrames into two parts based on a specified expression.
"""
from __future__ import annotations

from .base_splitter import BaseSplitter
from .expression_splitter import ExpressionSplitter
from .random_splitter import RandomSplitter


__all__ = ["BaseSplitter", "RandomSplitter", "ExpressionSplitter"]



"""This module contains the CacheHandler for serializing and deserializing data using pickle."""
import pickle
from pathlib import Path
from typing import Any

from .base_cache_handler import CacheHandler


def write_pickle(cache_file_path: Path, output: Any) -> None:
    """Writes the given data to a cache file at the specified path, serializing it using pickle.

    Args:
        cache_file_path: A Path object representing the location where the cache file should be saved.
        output: The data to be serialized and saved to the cache file.
    """
    with open(cache_file_path, "wb") as cache_file:
        pickle.dump(output, cache_file)


def read_pickle(cache_file_path: Path) -> Any:
    """Reads the cache file from the specified path and returns the deserialized data.

    Args:
        cache_file_path: A Path object representing the location of the cache file.

    Returns:
        The deserialized data stored in the cache file.
    """
    with open(cache_file_path, "rb") as cache_file:
        return pickle.load(cache_file)


PICKLE_CACHE_HANDLER = CacheHandler(write_pickle, read_pickle, "pkl")
"""A CacheHandler for pickling Python objects."""



"""Subpackage with utility functions and classes.

This subpackage contains utility functions and classes that are used throughout the project. These include a custom
logger, decorators, and helper functions for working with `vaex` DataFrames and `tqdm` progress bars.
"""
from __future__ import annotations

from .custom_logger import CustomLogger
from .decorators import auto_repr, timing
from .file_helpers import clear_directory
from .tqdm_helpers import set_tqdm_percent_wrapper
from .vaex_helpers import get_column, get_columns, get_filtered_df, get_indices


__all__ = [
    "CustomLogger",
    "auto_repr",
    "timing",
    "clear_directory",
    "set_tqdm_percent_wrapper",
    "get_column",
    "get_columns",
    "get_filtered_df",
    "get_indices",
]



"""The subpackage provides hyperparameter tuning functionality.

The following hyperparameter tuners are provided by the subpackage:
    - `BaseTuner`: The abstract base class for all hyperparameter tuners.
    - `OptunaTuner`: A hyperparameter tuner that uses Optuna for hyperparameter tuning.
"""
from .base_tuner import BaseTuner
from .optuna_tuner import OptunaTuner


__all__ = ["BaseTuner", "OptunaTuner"]



"""Module containing the TuneStep class."""
from __future__ import annotations

from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.model.tune.base_tuner import BaseTuner
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class TuneStep(PipelineStep):
    """Pipeline step for hyperparameter tuning."""

    _num_inputs = 2
    """Number of inputs expected by the TuneStep."""

    _num_outputs = 3
    """Number of outputs expected by the TuneStep."""

    @auto_repr
    def __init__(
        self,
        tuner: BaseTuner,
        inputs: list[str] | tuple[str, ...] | tuple[()] = (),
        outputs: list[str] | tuple[str, ...] | tuple[()] = (),
        cache_group: str | None = None,
    ) -> None:
        """Initialize the TuneStep with the specified tuner.

        Args:
            tuner: The tuner used for hyperparameter tuning.
            inputs: List or tuple of input keys expected by this step. Should contain two keys, corresponding to the
                DataSchema and the DataFrame to be used for hyperparameter tuning. The two inputs are passed to the
                tunes's objective function.
            outputs: List or tuple of output keys produced by this step. Should contain three keys, corresponding to
                the best hyperparameters, the best objective score, and an optional metadata object.
            cache_group: The cache group to use.
        """
        super().__init__(inputs, outputs, cache_group)
        self._tuner = tuner

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform hyperparameter tuning.

        Args:
            data_container: Contains the input.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            ValueError: If data container contains invalid data - not a vaex DataFrame.

        Returns:
            A DataContainer containing the output of the tuning step. It contains the best hyperparameters, the best
            objective score, and an optional metadata object.
        """
        data_schema = data_container.data[self._inputs[0]]
        if not isinstance(data_schema, DataSchema):
            raise ValueError(f"Invalid data type: {type(data_schema)}. Expected DataSchema.")

        dataframe = data_container.data[self._inputs[1]]
        if not isinstance(dataframe, DataFrame):
            raise ValueError(f"Invalid data type: {type(dataframe)}. Expected vaex DataFrame.")

        hyperparameters, score, metadata = self._tuner.tune(data_schema, dataframe, self._cache_group, force_recompute)
        data_container.data[self._outputs[0]] = hyperparameters
        data_container.data[self._outputs[1]] = score
        data_container.data[self._outputs[2]] = metadata
        return data_container



"""The subpackage provides functionality for fetching data from various sources.

This subpackage contains classes designed to easily fetch data from different sources, like AWS S3 or Kaggle,
and store them locally in specified destination directories. The main classes are 'BaseIngester', an abstract base
class for implementing specific data source classes, along with concrete implementations.

The following ingester classes are provided by the subpackage:
    - `BaseIngester`: The abstract base class for all ingesters.
        - `LocalFileEntry`: A class representing a local file entry.
        - `LocalManifest`: A class representing a local manifest.
        - `LocalManifestHandler`: A class for handling local manifests.
    - `S3Ingester`: An ingester for fetching data from AWS S3.
    - `KaggleIngester`: An ingester for fetching data from Kaggle.
"""
from __future__ import annotations

from .base_ingester import BaseIngester, LocalFileEntry, LocalManifest, LocalManifestHandler
from .kaggle_ingester import KaggleIngester
from .s3_ingester import S3Ingester


__all__ = ["BaseIngester", "S3Ingester", "KaggleIngester", "LocalFileEntry", "LocalManifest", "LocalManifestHandler"]



"""A module for downloading and managing Kaggle datasets using the Kaggle API.

In order to use this module, the user must have valid Kaggle API credentials.
"""
from __future__ import annotations

import hashlib
import json
import os
import shutil
from concurrent import futures
from dataclasses import dataclass
from datetime import datetime
from fnmatch import fnmatch
from itertools import repeat
from pathlib import Path
from typing import NamedTuple

import requests
from requests.auth import HTTPBasicAuth
from tqdm.auto import tqdm

from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_ingester import BaseIngester, LocalFileEntry, LocalManifestHandler


logger = CustomLogger()
"""A module-level custom logger."""


class KaggleCredentials(NamedTuple):
    """Represents a set of Kaggle API credentials, including a username and API key."""

    username: str
    """Username for the Kaggle API."""

    key: str
    """API key for the Kaggle API."""


class KaggleCredentialsManager:
    """Manages retrieval of Kaggle API credentials from environment variables or a file."""

    _ENV_VARIABLE_USERNAME = "KAGGLE_USERNAME"
    """Name of the environment variable containing the Kaggle username."""

    _ENV_VARIABLE_KEY = "KAGGLE_KEY"
    """Name of the environment variable containing the Kaggle API key."""

    _CONFIG_DEFAULT_PATH = "~/.kaggle/kaggle.json"
    """Default path to the Kaggle API credentials file."""

    _CONFIG_VARIABLE_USERNAME = "username"
    """Name of the key in the Kaggle API credentials file containing the Kaggle username."""

    _CONFIG_VARIABLE_KEY = "key"
    """Name of the key in the Kaggle API credentials file containing the Kaggle API key."""

    @staticmethod
    def get_kaggle_credentials(credentials_file_path: str | Path | None = None) -> KaggleCredentials:
        """Retrieves Kaggle API credentials from the specified file, environment variables, or the default location.

        Args:
            credentials_file_path: Path to the Kaggle API credentials file.

        Returns:
            A KaggleCredentials instance with the retrieved username and API key.
        """
        if credentials_file_path is not None:
            logger.info(f"Attempting to fetch Kaggle API credentials from config at {credentials_file_path}.")
            credentials = KaggleCredentialsManager._read_config_file(Path(credentials_file_path))
        else:
            logger.info(
                "Attempting to fetch Kaggle API credentials from environment variables "
                f"{KaggleCredentialsManager._ENV_VARIABLE_USERNAME!r} and "
                f"{KaggleCredentialsManager._ENV_VARIABLE_KEY!r}."
            )
            env_credentials = KaggleCredentialsManager._read_environment_config()

            if env_credentials is None:
                logger.info(
                    "Kaggle API credentials not found in environment variables, attempting to fetch from "
                    f"fallback path at {KaggleCredentialsManager._CONFIG_DEFAULT_PATH}."
                )
                credentials = KaggleCredentialsManager._read_config_file(
                    Path(KaggleCredentialsManager._CONFIG_DEFAULT_PATH)
                )
            else:
                credentials = env_credentials
        logger.info("Kaggle credentials successfully fetched.")

        return credentials

    @staticmethod
    def _read_config_file(credentials_file_path: Path) -> KaggleCredentials:
        """Reads Kaggle API credentials from the given configuration file.

        Args:
            credentials_file_path: Path to the Kaggle API credentials file.

        Returns:
            A KaggleCredentials instance with the retrieved username and API key.

        Raises:
            FileNotFoundError: If the file does not exist or is a directory.
            key_error: If the username or API key is missing from the credentials JSON.
            JSONDecodeError: If the JSON decoding fails while reading the credentials.
        """
        expanded_credentials_file_path = credentials_file_path.expanduser()
        if not expanded_credentials_file_path.exists():
            error_msg = f"{credentials_file_path} does not exist."
            logger.error(error_msg)
            raise FileNotFoundError(error_msg)

        if not expanded_credentials_file_path.is_file():
            error_msg = f"{credentials_file_path} is a directory."
            logger.error(error_msg)
            raise FileNotFoundError(error_msg)

        permissions = expanded_credentials_file_path.stat().st_mode
        if (permissions & 4) or (permissions & 32):
            logger.warning(
                "Your Kaggle API credentials are readable by other users on this system. "
                f"Run `chmod 600 {credentials_file_path}` to fix this issue."
            )

        try:
            with open(expanded_credentials_file_path) as f:
                config_data: dict[str, str] = json.load(f)
                kaggle_config = KaggleCredentials(
                    username=config_data[KaggleCredentialsManager._CONFIG_VARIABLE_USERNAME],
                    key=config_data[KaggleCredentialsManager._CONFIG_VARIABLE_KEY],
                )
                return kaggle_config
        except json.JSONDecodeError as e:
            logger.error(
                f"Failed to decode Kaggle API credentials JSON. {e.msg}: line "
                f"{e.lineno} column {e.colno} (char {e.pos})."
            )
            raise json.JSONDecodeError(e.msg, e.doc, e.pos) from e
        except KeyError as key_error:
            logger.error(f"Missing {key_error} from Kaggle API credentials JSON.")
            raise key_error

    @staticmethod
    def _read_environment_config() -> KaggleCredentials | None:
        """Reads Kaggle API credentials from environment variables.

        Returns:
            A KaggleCredentials instance with the retrieved username and API key, or None
                if the environment variables are not set.
        """
        username = os.getenv(KaggleCredentialsManager._ENV_VARIABLE_USERNAME)
        key = os.getenv(KaggleCredentialsManager._ENV_VARIABLE_KEY)
        return KaggleCredentials(username=username, key=key) if username and key else None


@dataclass
class KaggleFileManifest:
    """Manifest entry for a single file in a Kaggle dataset."""

    name: str
    """Name of the file."""

    creation_timestamp: float
    """Timestamp of the file creation."""

    total_bytes: int
    """Total size of the file in bytes."""


class KaggleIngester(BaseIngester):
    """Handles dataset retrieval from Kaggle, downloading and updating files as necessary.

    The `KaggleIngester` class downloads files from the specified Kaggle dataset and saves them to the destination
    directory. It also checks if the local files are up to date and skips downloading if everything is already in
    place.
    """

    _KAGGLE_API_VERSION = "v1"
    """The Kaggle API version to use."""

    _KAGGLE_DATASET_URL = f"https://www.kaggle.com/api/{_KAGGLE_API_VERSION}/datasets"
    """The base URL for Kaggle dataset API requests."""

    @auto_repr
    def __init__(
        self,
        owner_slug: str,
        dataset_slug: str,
        file_pattern: str | list[str] = "*",
        dataset_id: str | None = None,
        destination_directory: str | Path = "data/ingest-kaggle",
        dataset_version: str | int | None = None,
        kaggle_api_credentials_file: str | Path | None = None,
        num_workers: int = 64,
    ) -> None:
        """Initializes a `KaggleIngester` instance to fetch data from a specific Kaggle dataset.

        In order to use `KaggleIngester`, valid Kaggle API credentials are required. These credentials can be obtained
        by creating an API token on the Kaggle account settings page. The token should be saved in a JSON file named
        `kaggle.json` containing the "username" and "key" fields.

        There are three possible locations where Kaggle API credentials can be provided:

        1. Custom file location: Pass the file path to `kaggle_api_credentials_file` in the constructor.
        2. Environment variables: Set the KAGGLE_USERNAME and KAGGLE_KEY environment variables.
        3. Default .kaggle folder: Place the `kaggle.json` file into the "~/.kaggle/" directory.

        Note:
            The Kaggle API is not perfect and sometimes returns incorrect metadata for files, where one or more of the
            files are missing from the dataset. This can lead to the wrong files being downloaded or the download
            failing altogether. If you encounter this issue, please report it to Kaggle.

            The issue is observed when the dataset contains a large number of files (e.g. 1000+) or if the dataset
            contains nested folders.

        Args:
            owner_slug: The owner's Kaggle username or organization name.
            dataset_slug: The dataset's unique Kaggle identifier (slug).
            file_pattern: Pattern to match the files to download, e.g. `*.csv` or [`*.csv`, `*.json`], etc.
                For more information, see https://docs.python.org/3/library/fnmatch.html.
            dataset_id: Id of the dataset to be used instead of the default fingerprint (MD5 hash of the owner slug,
                dataset version, and dataset slug). Note that this will overwrite any existing dataset with the same
                name in the destination directory, so make sure to use a unique name.
            destination_directory: The directory where the downloaded files will be stored.
            dataset_version: The specific dataset version number to download. If not provided,
                the latest version will be fetched.
            kaggle_api_credentials_file: Path to a Kaggle API credentials JSON file. If not
                provided, environment variables or the default file location will be used.
            num_workers: Number of concurrent threads to use when downloading files.

        Examples:
            >>> from mleko.dataset.sources import KaggleIngester
            >>> kaggle_ingester = KaggleIngester(
            ...     owner_slug="allen-institute-for-ai",
            ...     dataset_slug="covid-19-masks-dataset",
            ...     file_pattern="file_*.zip",
            ...     dataset_id="covid-19", # Optional, but will store the data in "./data/covid-19/" instead of
            ...                            # "./data/<fingerprint>/".
            ...     destination_directory="data",
            ...     dataset_version=1,
            ... )
            >>> kaggle_ingester.fetch_data()
            [PosixPath('~/data/covid-19/file_1.zip'), PosixPath('~/data/covid-19/file_2.zip')]
        """
        dataset_id = (
            dataset_id
            if dataset_id is not None
            else hashlib.md5((owner_slug + dataset_slug + str(dataset_version)).encode()).hexdigest()
        )
        super().__init__(destination_directory, dataset_id)
        self._local_manifest_handler = LocalManifestHandler(
            self._destination_directory / f"{self._fingerprint}.manifest.json"
        )
        self._owner_slug = owner_slug
        self._dataset_slug = dataset_slug
        self._dataset_version = dataset_version
        self._kaggle_config = KaggleCredentialsManager.get_kaggle_credentials(kaggle_api_credentials_file)
        self._num_workers = num_workers

        if isinstance(file_pattern, str):
            file_pattern = [file_pattern]
        self._file_pattern = file_pattern

    def fetch_data(self, force_recompute: bool = False) -> list[Path]:
        """Fetches data from the specified Kaggle dataset.

        This method downloads files from the Kaggle dataset and returns the local file paths of the downloaded files.
        The method checks if local files are up-to-date and skips downloading if everything is already in place and
        `force_recompute` is set to False.

        Args:
            force_recompute: If set to False, the method will check if the local files are up-to-date and
                skip downloading if everything is already in place.

        Returns:
            A list of local file paths pointing to the downloaded files.

        Raises:
            ValueError: If Kaggle returns 0 files for the given dataset. This could occur if the API is broken.
        """
        dataset_path = f"{self._owner_slug}/{self._dataset_slug}"

        params: dict[str, str] = {}
        if self._dataset_version:
            params["datasetVersionNumber"] = str(self._dataset_version)
        kaggle_manifest = self._build_kaggle_manifest(params)

        if force_recompute:
            logger.info(
                f"\033[33mForce Cache Refresh\033[0m: Downloading files matching {self._file_pattern} from "
                f"{dataset_path} to {self._destination_directory} from Kaggle."
            )
        else:
            if self._is_local_dataset_fresh(kaggle_manifest):
                logger.info("\033[32mCache Hit\033[0m: Local dataset is up to date with Kaggle, skipping download.")
                local_file_names = set(self._local_manifest_handler.get_file_names())
                kaggle_file_names: set[str] = {kaggle_file.name for kaggle_file in kaggle_manifest}
                files_to_delete = list(local_file_names.difference(kaggle_file_names))

                if len(files_to_delete) > 0:
                    logger.info(
                        f"Deleting {len(files_to_delete)} files from "
                        f"{self._destination_directory} that are no longer present in Kaggle or filtered out."
                    )

                self._delete_local_files(files_to_delete)
                self._local_manifest_handler.remove_files(files_to_delete)
                return self._get_full_file_paths(self._local_manifest_handler.get_file_names())

            logger.info(
                f"\033[31mCache Miss\033[0m: Downloading files matching {self._file_pattern} from "
                f"{dataset_path} to {self._destination_directory} from Kaggle."
            )

        self._delete_local_files(self._local_manifest_handler.get_file_names())
        kaggle_file_paths = [f"{dataset_path}/{file_metadata.name}" for file_metadata in kaggle_manifest]
        if len(kaggle_file_paths) > 0:
            self._kaggle_fetch_files(kaggle_file_paths, params)
            self._local_manifest_handler.set_files(
                [
                    LocalFileEntry(
                        name=Path(kaggle_file_paths[0]).name,
                        size=os.path.getsize(self._destination_directory / Path(kaggle_file_paths[0]).name),
                    )
                ]
            )
            logger.info(f"Finished downloading {len(kaggle_file_paths)} files from Kaggle.")

        return self._get_full_file_paths(self._local_manifest_handler.get_file_names())

    def _build_kaggle_manifest(self, params: dict[str, str]) -> list[KaggleFileManifest]:
        """Fetch the metadata of the files in the dataset.

        When fetching the metadata, the API returns a list of files in the dataset. The list contains the name of the
        file, the creation date, and the file size.

        Args:
            params: A dictionary of query parameters to pass to the Kaggle API.

        Raises:
            HTTPError: If there is an error in the HTTP response while requesting file list from Kaggle.

        Returns:
            A list of KaggleFileManifest objects containing the metadata of the files in the dataset.
        """
        list_files_response = requests.get(
            f"{self._KAGGLE_DATASET_URL}/list/{self._owner_slug}/{self._dataset_slug}",
            params=params,
            auth=HTTPBasicAuth(self._kaggle_config.username, self._kaggle_config.key),
            timeout=5,
        )

        try:
            list_files_response.raise_for_status()
        except requests.HTTPError as e:
            logger.error(e)
            raise requests.HTTPError(e) from e

        kaggle_manifest: list[KaggleFileManifest] = [
            KaggleFileManifest(
                file["name"],
                datetime.strptime(file["creationDate"].split(".")[0], "%Y-%m-%dT%H:%M:%S").timestamp(),
                int(file["totalBytes"]),
            )
            for file in json.loads(list_files_response.content)["datasetFiles"]
            if any(fnmatch(file["name"], pattern) for pattern in self._file_pattern)
        ]

        if len(kaggle_manifest) == 0:
            error_msg = (
                f"No files matching {self._file_pattern} found in Kaggle dataset "
                f"{self._owner_slug}/{self._dataset_slug}."
            )
            logger.error(error_msg)
            raise FileNotFoundError(error_msg)
        logger.info(f"Found {len(kaggle_manifest)} file(s) matching any of {self._file_pattern} in Kaggle dataset.")

        return kaggle_manifest

    def _kaggle_fetch_file(self, kaggle_file_path: str, params: dict[str, str]) -> None:
        """Downloads a single Kaggle dataset file and saves it in the destination directory.

        Args:
            kaggle_file_path: The Kaggle file path to download.
            params: The request parameters containing the dataset version number, if applicable.

        Raises:
            HTTPError: If there is an error in the HTTP response while downloading file from Kaggle.
        """
        local_file_path = self._destination_directory / Path(kaggle_file_path).name
        response = requests.get(
            f"{self._KAGGLE_DATASET_URL}/download/{kaggle_file_path}",
            params=params,
            auth=HTTPBasicAuth(self._kaggle_config.username, self._kaggle_config.key),
            timeout=5,
            stream=True,
        )

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            logger.error(e)
            raise requests.HTTPError(e) from e

        with open(local_file_path, "wb") as f:
            shutil.copyfileobj(response.raw, f)

        file_signature = None
        with open(local_file_path, "rb") as f:
            file_signature = f.read(4)

        # Check if file is a zip file and unzip it if it is.
        if file_signature == b"PK\x03\x04":
            local_file_path = local_file_path.rename(local_file_path.with_name(local_file_path.name + ".zip"))
            shutil.unpack_archive(local_file_path, local_file_path.parent)
            local_file_path.unlink()

    def _kaggle_fetch_files(self, kaggle_file_paths: list[str], params: dict[str, str]) -> None:
        """Downloads multiple Kaggle dataset files concurrently.

        Args:
            kaggle_file_paths: A list of Kaggle file paths to download.
            params: The request parameters containing the dataset version number, if applicable.
        """
        with tqdm(total=len(kaggle_file_paths), desc="Downloading files from Kaggle") as pbar:
            with futures.ThreadPoolExecutor(max_workers=min(self._num_workers, len(kaggle_file_paths))) as executor:
                for _ in executor.map(
                    self._kaggle_fetch_file,
                    kaggle_file_paths,
                    repeat(params),
                ):
                    pbar.update(1)

    def _is_local_dataset_fresh(self, files_metadata: list[KaggleFileManifest]) -> bool:
        """Checks if the local dataset files are up to date with the Kaggle dataset files.

        Comparing file size and modification timestamp, this method determines if the local files are up to date and
        if they match the remote Kaggle dataset files.

        Args:
            files_metadata: A list containing the metadata of the files in the Kaggle dataset.

        Returns:
            True if the local dataset files are up to date, False otherwise.
        """
        for file in files_metadata:
            local_file_path = self._destination_directory / file.name
            if (
                not local_file_path.exists()
                or file.total_bytes != os.path.getsize(local_file_path)
                or file.creation_timestamp > os.path.getmtime(local_file_path)
            ):
                return False
        return True



"""Module handling data ingestion from a data source in the pipeline process.

This module contains the `IngestStep` class which is a specialized pipeline step designed for handling data
fetching from a specified `BaseIngester`. It's responsible for retrieving data from the data source, and
returning a `DataContainer` object containing the list of fetched files.
"""
from __future__ import annotations

from mleko.dataset.ingest import BaseIngester
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils import auto_repr


class IngestStep(PipelineStep):
    """Pipeline step that manages data ingestion from a configured data source."""

    _num_inputs = 0
    """Number of inputs expected by the IngestStep."""

    _num_outputs = 1
    """Number of outputs expected by the IngestStep."""

    @auto_repr
    def __init__(
        self,
        ingester: BaseIngester,
        inputs: list[str] | tuple[str, ...] | tuple[()] = (),
        outputs: list[str] | tuple[str, ...] | tuple[()] = (),
    ) -> None:
        """Initialize the IngestStep with the specified data source.

        Args:
            ingester: The data source from which to fetch the data, a BaseIngester instance.
            inputs: List or tuple of input keys expected by this step. Should be empty.
            outputs: List or tuple of output keys produced by this step. Should contain a single key,
                corresponding to the list of fetched file Paths.
        """
        super().__init__(inputs, outputs, None)
        self._ingester = ingester

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Fetch data from the configured data source and return a DataContainer with fetched files.

        Args:
            data_container: Input data for this step's processing operation.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Returns:
            DataContainer: A DataContainer containing a list of fetched files.
        """
        files = self._ingester.fetch_data(force_recompute)
        data_container.data[self._outputs[0]] = files
        return data_container



"""Module handling dataframe splitting.

This module contains the SplitStep class, which is responsible for splitting a DataFrame into two parts.
It uses a BaseDataSplitter to perform the actual splitting.
"""
from __future__ import annotations

from vaex import DataFrame

from mleko.dataset.split.base_splitter import BaseSplitter
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class SplitStep(PipelineStep):
    """Pipeline step that splits a DataFrame into two parts."""

    _num_inputs = 1
    """Number of inputs expected by the SplitStep."""

    _num_outputs = 2
    """Number of outputs expected by the SplitStep."""

    @auto_repr
    def __init__(
        self,
        splitter: BaseSplitter,
        inputs: list[str] | tuple[str, ...] | tuple[()] = (),
        outputs: list[str] | tuple[str, ...] | tuple[()] = (),
        cache_group: str | None = None,
    ) -> None:
        """Initialize the SplitStep with the specified data splitter.

        Args:
            splitter: The DataSplitter responsible for handling data splitting.
            inputs: List or tuple of input keys expected by this step. Should contain a single key,
                corresponding to the DataFrame to be split.
            outputs: List or tuple of output keys produced by this step. Should contain two keys,
                corresponding to the two DataFrames after splitting.
            cache_group: The cache group to use.
        """
        super().__init__(inputs, outputs, cache_group)
        self._splitter = splitter

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform data splitting using the configured splitter.

        Args:
            data_container: Contains the DataFrame to be split.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            ValueError: If data container contains invalid data - not a vaex DataFrame.

        Returns:
            A DataContainer containing the split data as two vaex DataFrames.
        """
        dataframe = data_container.data[self._inputs[0]]
        if not isinstance(dataframe, DataFrame):
            raise ValueError(f"Invalid data type: {type(dataframe)}. Expected vaex DataFrame.")

        df1, df2 = self._splitter.split(dataframe, self._cache_group, force_recompute)
        data_container.data[self._outputs[0]] = df1
        data_container.data[self._outputs[1]] = df2
        return data_container



"""Abstract base class module for data source implementations to fetch and store data from various sources."""
from __future__ import annotations

import dataclasses
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path

from mleko.utils.custom_logger import CustomLogger


logger = CustomLogger()
"""A module-level custom logger."""


@dataclass
class LocalFileEntry:
    """Manifest entry for a single local file."""

    name: str
    """Name of the file."""

    size: int
    """Size of the file in bytes."""


@dataclass
class LocalManifest:
    """Manifest for the local dataset."""

    files: list[LocalFileEntry]
    """List of files in the local dataset."""


class LocalManifestHandler:
    """`LocalManifestHandler` provides a convenient interface for reading and writing local manifest files."""

    def __init__(self, manifest_path: str | Path):
        """Initializes the local manifest handler.

        The manifest is intended to be used to keep track of the downloaded file names and sizes.
        It should reflect the current state of the local dataset.

        Args:
            manifest_path: Path to the manifest file.
        """
        self._manifest_path = Path(manifest_path)

    def set_files(self, file_data: list[LocalFileEntry]) -> None:
        """Sets the manifest to the specified files.

        Args:
            file_data: List of file data to set the manifest to.
        """
        manifest_data = self._read_manifest()
        manifest_data.files = file_data
        self._write_manifest(manifest_data)

    def remove_files(self, file_names: list[str]) -> None:
        """Removes the specified files from the manifest.

        Args:
            file_names: List of file names to remove from the manifest.
        """
        manifest_data = self._read_manifest()
        manifest_data.files = [file for file in manifest_data.files if file.name not in file_names]
        self._write_manifest(manifest_data)

    def get_file_names(self) -> list[str]:
        """Gets the list of file names in the manifest.

        Returns:
            List of file names in the manifest.
        """
        manifest = self._read_manifest()
        return [file.name for file in manifest.files]

    def _read_manifest(self) -> LocalManifest:
        """Reads the manifest from the manifest file.

        Returns:
            Manifest data read from the manifest file.
        """
        try:
            if self._manifest_path.exists():
                with open(self._manifest_path) as manifest_file:
                    manifest_dict = json.load(manifest_file)
                    local_manifest = self._deserialize_manifest(manifest_dict)
                    return local_manifest
        except (json.JSONDecodeError, TypeError, AttributeError) as e:
            logger.error(f"Error decoding manifest file {self._manifest_path}: {e}")
        return LocalManifest([])

    def _write_manifest(self, manifest_data: LocalManifest) -> None:
        """Writes the manifest to the manifest file.

        Args:
            manifest_data: Manifest data to write to the manifest file.
        """
        with open(self._manifest_path, "w") as manifest_file:
            json.dump(dataclasses.asdict(manifest_data), manifest_file, indent=2)

    def _deserialize_manifest(self, manifest_dict: dict) -> LocalManifest:
        """Deserializes the manifest from the manifest dictionary.

        Args:
            manifest_dict: Manifest dictionary to deserialize.

        Returns:
            Deserialized manifest.
        """
        files = [LocalFileEntry(**file_dict) for file_dict in manifest_dict.get("files", [])]
        return LocalManifest(files=files)


class BaseIngester(ABC):
    """`BaseIngester` is an abstract base class for data source classes that interact with various external sources.

    This class provides the basic structure and methods necessary for derived data source classes, facilitating data
    fetching from various sources.
    """

    def __init__(self, destination_directory: str | Path, fingerprint: str | None) -> None:
        """Initializes the data source and ensures the destination directory exists.

        Args:
            destination_directory: Directory where the fetched data will be stored locally.
            fingerprint: Optional fingerprint to append to the destination directory.
        """
        self._destination_directory = Path(destination_directory)
        if fingerprint is not None:
            self._fingerprint = fingerprint
            self._destination_directory = self._destination_directory / fingerprint
        self._destination_directory.mkdir(parents=True, exist_ok=True)

    @abstractmethod
    def fetch_data(self, force_recompute: bool = False) -> list[Path]:
        """Downloads and stores data in the 'destination_directory' using the specific data source implementation.

        Args:
            force_recompute: Whether to force the data source to recompute its output, even if it already exists.

        Raises:
            NotImplementedError: Must be implemented in the child class that inherits from `BaseIngester`.
        """
        raise NotImplementedError

    def _get_full_file_paths(self, file_names: list[str]) -> list[Path]:
        """Gets the full file paths for the specified file names.

        Note that this method only returns the file paths for files that exist locally.

        Args:
            file_names: List of file names to get the full file paths for.

        Returns:
            List of full file paths for the specified file names.
        """
        return [
            self._destination_directory / file_name
            for file_name in file_names
            if (self._destination_directory / file_name).exists()
        ]

    def _delete_local_files(self, file_names: list[str]) -> None:
        """Deletes the specified files from the local dataset.

        Args:
            file_names: List of file names to delete from the local dataset.
        """
        for file_path in self._get_full_file_paths(file_names):
            file_path.unlink()



"""This module contains helper functions for working with `vaex` DataFrames."""
from __future__ import annotations

import vaex


def get_column(df: vaex.DataFrame, column: str) -> vaex.Expression:
    """Get specified column from a DataFrame as an Expression.

    Args:
        df: The input DataFrame.
        column: The name of the desired column.

    Returns:
        The specified column as an Expression.
    """
    return df[column]


def get_columns(df: vaex.DataFrame, columns: list[str]) -> vaex.DataFrame:
    """Get specified columns from a DataFrame.

    Args:
        df: The input DataFrame.
        columns: A list of the names of the desired columns or an Expression.

    Returns:
        A DataFrame containing only the specified columns.

    Examples:
        >>> import vaex
        >>> from mleko.utils import get_columns
        >>> df = vaex.from_arrays(column1=[1, 2, 3], column2=[4, 5, 6], column3=[7, 8, 9])
        >>> get_columns(df, ["column1", "column3"]).get_column_names()
        ['column1', 'column3']
    """
    return df[columns]


def get_filtered_df(df: vaex.DataFrame, filter: vaex.Expression) -> vaex.DataFrame:
    """Get filtered DataFrame.

    Will return a DataFrame containing only the rows that satisfy the filter. The filter is an Expression that
    evaluates to a boolean value for each row.

    Args:
        df: The input DataFrame.
        filter: A boolean Expression used to filter the DataFrame.

    Returns:
        A DataFrame containing only the rows that satisfy the filter.

    Examples:
        >>> import vaex
        >>> from mleko.utils import get_filtered_df
        >>> df = vaex.from_arrays(column1=[1, 2, 3], column2=[4, 5, 6], column3=[7, 8, 9])
        >>> get_filtered_df(df, df.column1 > 1)
        #    column1    column2    column3
        0          2          5          8
        1          3          6          9
    """
    return df[filter]


def get_indices(df: vaex.DataFrame, indices: list[int]) -> vaex.DataFrame:
    """Get DataFrame containing only the specified indices.

    Args:
        df: The input DataFrame.
        indices: A list of the indices to be extracted.

    Returns:
        A DataFrame containing only the specified indices.

    Examples:
        >>> import vaex
        >>> from mleko.utils import get_indices
        >>> df = vaex.from_arrays(column1=[1, 2, 3], column2=[4, 5, 6], column3=[7, 8, 9])
        >>> get_indices(df, [0, 2])
        #    column1    column2    column3
        0          1          4          7
        1          3          6          9
    """
    idx_name = "index"
    df[idx_name] = vaex.vrange(0, df.shape[0])
    index = get_column(df, idx_name)
    selection = get_filtered_df(df, index.isin(indices))
    selection.delete_virtual_column(idx_name)
    return selection.extract()



"""The module containing the VaexFingerprinter class."""
from __future__ import annotations

import hashlib

import vaex

from .base_fingerprinter import BaseFingerprinter


class VaexFingerprinter(BaseFingerprinter):
    """A fingerprinter for Vaex DataFrames."""

    def fingerprint(self, data: vaex.DataFrame) -> str:
        """Generate a fingerprint for a Vaex DataFrame.

        Note:
            The fingerprint is generated by calling the `fingerprint()` method of the `vaex` DataFrame. This
            method will attempt to generate a unique fingerprint for the given DataFrame, it may however
            not be completely stable across different runs of the program (see the `fingerprint()` method
            of the `vaex` DataFrame for more information).

        Args:
            data: The Vaex DataFrame to be fingerprinted.

        Returns:
            The fingerprint as a hexadecimal string.

        Examples:
            >>> fingerprinter = VaexFingerprinter()
            >>> fingerprinter.fingerprint(vaex.from_arrays(x=[1, 2, 3], y=[4, 5, 6]))
            "fingerprint"
        """
        fingerprint = hashlib.md5(str(data.fingerprint()).encode()).hexdigest()
        return fingerprint



"""The subpackage provides caching mixins and fingerprinting utilities to efficiently cache method call results.

The core of the caching functionality is provided by the `CacheMixin` class, which can be used to
cache the results of method calls.

The following caching mixins are provided by the subpackage:
    - `CacheMixin`: The core caching mixin.
    - `LRUCacheMixin`: A mixin that adds LRU eviction to the cache.
"""
from __future__ import annotations

from .cache_mixin import CacheMixin
from .lru_cache_mixin import LRUCacheMixin


__all__ = ["CacheMixin", "LRUCacheMixin"]



"""Pipeline subpackage for managing and executing data processing steps.

This subpackage provides the necessary components to create a customizable data processing pipeline. It includes
abstract base classes for pipeline steps, a flexible and customizable Pipeline class for managing and executing a
series of data processing steps, and a subpackage named `steps` that contains concrete implementations of the
`PipelineStep` class.
"""
from __future__ import annotations

from .data_container import DataContainer
from .pipeline import Pipeline
from .pipeline_step import PipelineStep


__all__ = ["Pipeline", "PipelineStep", "DataContainer"]



"""The subpackage provides functionality for selecting features from data.

This subpackage offers a collection of feature selectors, each designed for a specific type of feature selection
task. By using these unique feature selectors sequentially, you can create a complete feature selection workflow
within the pipeline.

The following feature selectors are provided by the subpackage:
    - `BaseFeatureSelector`: The abstract base class for all feature selectors.
    - `CompositeFeatureSelector`: A feature selector for combining multiple feature selectors.
    - `MissingRateFeatureSelector`: A feature selector for removing features with a high percentage of missing values.
    - `VarianceFeatureSelector`: A feature selector for removing features with a low standard deviation.
    - `PearsonCorrelationFeatureSelector`: A feature selector for removing features with a high Pearson correlation.
    - `InvarianceFeatureSelector`: A feature selector for removing invariant features.
"""
from .base_feature_selector import BaseFeatureSelector
from .composite_feature_selector import CompositeFeatureSelector
from .invariance_feature_selector import InvarianceFeatureSelector
from .missing_rate_feature_selector import MissingRateFeatureSelector
from .pearson_correlation_feature_selector import PearsonCorrelationFeatureSelector
from .variance_feature_selector import VarianceFeatureSelector


__all__ = [
    "BaseFeatureSelector",
    "CompositeFeatureSelector",
    "MissingRateFeatureSelector",
    "VarianceFeatureSelector",
    "PearsonCorrelationFeatureSelector",
    "InvarianceFeatureSelector",
]



"""Module containing the TransformStep class."""
from __future__ import annotations

from typing import Literal

from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.dataset.transform.base_transformer import BaseTransformer
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class TransformStep(PipelineStep):
    """Pipeline step for transformation of features in DataFrame."""

    _num_inputs = 2
    """Number of inputs expected by the TransformStep."""

    _num_outputs = 2
    """Number of outputs expected by the TransformStep."""

    @auto_repr
    def __init__(
        self,
        transformer: BaseTransformer,
        action: Literal["fit", "transform", "fit_transform"],
        inputs: list[str] | tuple[str, ...] | tuple[()] = (),
        outputs: list[str] | tuple[str, ...] | tuple[()] = (),
        cache_group: str | None = None,
    ) -> None:
        """Initialize the TransformStep with the specified transformer.

        The action parameter specifies whether the transformer should be fitted, transformed, or both. If the action
        is "fit" the `TransformStep` will return the fitted transformer. If the action is "transform" the
        `TransformStep` will return the transformed DataFrame. If the action is "fit_transform" the `TransformStep`
        will return the updated data schema, the fitted transformer and the transformed DataFrame in the
        specified order.

        Args:
            transformer: The Transformer responsible for handling feature transformation.
            action: The action to perform, one of "fit", "transform", or "fit_transform".
            inputs: List or tuple of input keys expected by this step. Should contain two keys,
                corresponding to the data schema of the DataFrame and the DataFrame to be transformed.
            outputs: List or tuple of output keys produced by this step. If the action is "fit" or "transform",
                should contain a two keys, corresponding to the updated data schema and a fitted transformer or the
                transformed DataFrame. If the action is "fit_transform", should contain three keys, corresponding
                to the updated data schema, fitted transformer and the transformed DataFrame.
            cache_group: The cache group to use.

        Raises:
            ValueError: If action is not one of "fit", "transform", or "fit_transform".
        """
        if action not in ("fit", "transform", "fit_transform"):
            raise ValueError(
                f"Invalid action: {action}. Expected one of 'fit', 'transform', or 'fit_transform'."
            )  # pragma: no cover

        if action == "fit_transform":
            self._num_outputs = 3

        super().__init__(inputs, outputs, cache_group)
        self._transformer = transformer
        self._action = action

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform transformation using the configured transformer.

        Args:
            data_container: Contains the input DataFrame.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            ValueError: If data container contains invalid data - not a vaex DataFrame.

        Returns:
            A DataContainer containing the output of the action performed by the step, the updated data schema, and
            either the fitted transformer, the transformed DataFrame, or all.
        """
        data_schema = data_container.data[self._inputs[0]]
        if not isinstance(data_schema, DataSchema):
            raise ValueError(f"Invalid data type: {type(data_schema)}. Expected DataSchema.")

        dataframe = data_container.data[self._inputs[1]]
        if not isinstance(dataframe, DataFrame):
            raise ValueError(f"Invalid data type: {type(dataframe)}. Expected vaex DataFrame.")

        if self._action == "fit":
            ds, transformer = self._transformer.fit(data_schema, dataframe, self._cache_group, force_recompute)
            data_container.data[self._outputs[0]] = ds
            data_container.data[self._outputs[1]] = transformer
        elif self._action == "transform":
            ds, df = self._transformer.transform(data_schema, dataframe, self._cache_group, force_recompute)
            data_container.data[self._outputs[0]] = ds
            data_container.data[self._outputs[1]] = df
        elif self._action == "fit_transform":
            ds, transformer, df = self._transformer.fit_transform(
                data_schema, dataframe, self._cache_group, force_recompute
            )
            data_container.data[self._outputs[0]] = ds
            data_container.data[self._outputs[1]] = transformer
            data_container.data[self._outputs[2]] = df
        return data_container



"""Module for the base hyperparameter tuning class."""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Callable, Hashable, Literal

import optuna
import vaex
from optuna.samplers import (
    BaseSampler,
    BruteForceSampler,
    CmaEsSampler,
    GridSampler,
    NSGAIIISampler,
    NSGAIISampler,
    PartialFixedSampler,
    QMCSampler,
    RandomSampler,
    TPESampler,
)
from optuna.samplers.nsgaii._child_generation_strategy import NSGAIIChildGenerationStrategy
from tqdm.auto import tqdm

from mleko.cache.fingerprinters import (
    CallableSourceFingerprinter,
    OptunaPrunerFingerprinter,
    OptunaSamplerFingerprinter,
)
from mleko.dataset.data_schema import DataSchema
from mleko.model.base_model import HyperparametersType
from mleko.utils.custom_logger import CustomLogger

from .base_tuner import BaseTuner


logger = CustomLogger()
"""The logger for the module."""

OptimizeDirection = Literal["maximize", "minimize"]
"""Literal type for the direction of optimization."""


class OptunaTuner(BaseTuner):
    """Hyperparameter tuner using `Optuna`."""

    def __init__(
        self,
        cache_directory: str | Path,
        objective_function: Callable[
            [optuna.Trial, DataSchema, vaex.DataFrame], float | list[float] | tuple[float, ...]
        ],
        direction: OptimizeDirection | list[OptimizeDirection],
        num_trials: int,
        sampler: BaseSampler | None = None,
        pruner: optuna.pruners.BasePruner | None = None,
        random_state: int | None = None,
        cache_size: int = 1,
    ) -> None:
        """Initializes a new OptunaTuner instance.

        Note:
            For more information about Optuna, please refer to the documentation:
            https://optuna.readthedocs.io/en/stable/

        Warning:
            The caching functionality of the obective function is implemented by
            serializing the function source code itself. Ensure that all dependencies
            of the objective function are defined within the function itself. Otherwise,
            the dependencies will not be included in the fingerprint of the tuner and
            the results of the hyperparameter tuning will be unpredictable. For example,
            if the objective function depends on a global variable, the cahing
            functionality will not detect changes to the value itself and will not
            recompute the result.

            In addition, the objective function should preferably not use any cached
            methods, such as `BaseModel.fit_transform`. Instead, the objective
            function should use the underscored methods (`BaseModell._fit_transform`)
            to avoid caching the results of each trial.

        Args:
            cache_directory: The target directory where the output is to be saved.
            objective_function: The objective function to optimize. The function must
                accept three arguments: the Optuna trial, the data schema, and the
                DataFrame to be tuned on. The function must return either a single
                float value or a list/tuple of float values. If a list/tuple is
                returned, the tuner will perform multi-objective optimization.
            direction: The direction of optimization. Either "maximize" or "minimize".
                If a list of directions is given, the tuner will perform multi-objective
                optimization. The length of the list must match the length of the list
                returned by the objective function.
            num_trials: The number of trials to perform.
            sampler: The Optuna sampler to use, if None `TPESampler` is
                used for single-objective optimization and `NSGAIISampler`
                is used for multi-objective optimization.
            pruner: The Optuna pruner to use, if None `optuna.pruners.MedianPruner` is
                used.
            random_state: The random state to use for the Optuna sampler. If None, the
                default random state of the sampler is used. Setting this will override
                the random state of the sampler.
            cache_size: The maximum number of cache entries.

        Examples:
            >>> import vaex
            >>> from mleko.model import LGBMModel
            >>> from mleko.tune import OptunaTuner
            >>> from mleko.dataset import DataSchema
            >>> def objective_function(trial, data_schema, dataframe):
            ...     params = {
            ...         "num_iterations": trial.suggest_int("num_iterations", 10, 100),
            ...         "learning_rate": trial.suggest_float("learning_rate", 0.01, 0.1),
            ...         "num_leaves": trial.suggest_int("num_leaves", 10, 100),
            ...     }
            >>>     model = LGBMModel(
            ...         cache_directory=".",
            ...         target="class_",
            ...         features=["sepal_width", "petal_length", "petal_width"],
            ...         num_iterations=100,
            ...         learning_rate=0.1,
            ...         num_leaves=31,
            ...         random_state=42,
            ...         metric=["auc"],
            ...     )
            >>>     df_train, df_val = dataframe.ml.random_split(test_size=0.20, verbose=False)
            >>>     _, metrics, _, _ = model._fit_transform(data_schema, df_train, df_val, params)
            >>>     return metrics['validation']['auc'][-1]
            >>> optuna_tuner = OptunaTuner(
            ...     cache_directory=f"data/{DATASET_NAME}/tune",
            ...     objective_function=objective_function,
            ...     direction="maximize",
            ...     num_trials=51,
            ...     random_state=RANDOM_STATE,
            ... )
            >>> dataframe = vaex.ml.datasets.load_iris()
            >>> data_schema = DataSchema(
            ...     numerical=["sepal_length", "sepal_width", "petal_length", "petal_width"],
            ... )
            >>> best_trial, best_score, study = optuna_tuner.tune(data_schema, dataframe)
        """
        super().__init__(cache_directory, cache_size)
        self._objective_function = objective_function
        self._direction = direction
        self._num_trials = num_trials
        self._sampler = sampler or (TPESampler() if isinstance(direction, list) else NSGAIISampler())
        self._pruner = pruner or optuna.pruners.MedianPruner()
        self._random_state = random_state

        self._reset_sampler_rng(self._sampler)

        optuna_logger = logging.getLogger("optuna")
        for handler in optuna_logger.handlers:
            optuna_logger.removeHandler(handler)
        for handler in logger.handlers:
            optuna_logger.addHandler(handler)

    def _tune(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame
    ) -> tuple[HyperparametersType, float | list[float] | tuple[float, ...], optuna.study.Study]:
        """Perform the hyperparameter tuning.

        Args:
            data_schema: Data schema for the DataFrame.
            dataframe: DataFrame to be tuned on.

        Returns:
            Tuple containing the best hyperparameters, the best score, and a the Optuna study.
        """
        self._reset_sampler_rng(self._sampler)

        if isinstance(self._direction, list):
            study = optuna.create_study(sampler=self._sampler, pruner=self._pruner, directions=self._direction)
        else:
            study = optuna.create_study(sampler=self._sampler, pruner=self._pruner, direction=self._direction)

        with tqdm(total=self._num_trials) as pbar:

            def tqdm_callback(study: optuna.study.Study, _trial: optuna.trial.FrozenTrial) -> None:
                pbar.update(1)
                best_trial = study.best_trials[0].number
                best_score = study.best_trials[0].values

                if isinstance(best_score, (list, tuple)) and len(best_score) == 1:
                    best_values = f"{best_score[0]:.4f}"
                else:
                    best_values = f"[{', '.join(f'{x:.4f}' for x in best_score)}]"

                pbar.set_description(f"Best trial: {best_trial} | Best score: {best_values}")

            study.optimize(
                lambda trial: self._objective_function(trial, data_schema.copy(), dataframe.copy()),
                n_trials=self._num_trials,
                callbacks=[tqdm_callback],
            )

        best_parameters = study.best_trials[0].params
        best_score = study.best_trials[0].values
        best_score = best_score[0] if isinstance(best_score, (list, tuple)) and len(best_score) == 1 else best_score

        self._reset_sampler_rng(self._sampler)

        return best_parameters, best_score, study

    def _fingerprint(self) -> Hashable:
        """Generates a fingerprint for the tuner.

        Returns:
            The fingerprint of the tuner.
        """
        return (
            super()._fingerprint(),
            CallableSourceFingerprinter().fingerprint(self._objective_function),
            self._direction,
            self._num_trials,
            OptunaSamplerFingerprinter().fingerprint(self._sampler),
            OptunaPrunerFingerprinter().fingerprint(self._pruner),
            self._random_state,
        )

    def _reset_sampler_rng(self, sampler: BaseSampler) -> None:
        """Resets the random number generator of the given Optuna sampler.

        Args:
            sampler: The Optuna sampler to reset the random number generator of.
        """
        if isinstance(
            sampler,
            (
                GridSampler,
                RandomSampler,
                TPESampler,
                NSGAIISampler,
                NSGAIIISampler,
                BruteForceSampler,
            ),
        ):
            sampler._rng.rng.seed(self._random_state)

        if isinstance(
            sampler,
            (
                TPESampler,
                NSGAIISampler,
                NSGAIIISampler,
            ),
        ):
            self._reset_sampler_rng(sampler._random_sampler)

        if isinstance(sampler, CmaEsSampler):
            sampler._cma_rng.rng.seed(self._random_state)

        if isinstance(sampler, (CmaEsSampler, QMCSampler)):
            self._reset_sampler_rng(sampler._independent_sampler)

        if isinstance(sampler, PartialFixedSampler):
            self._reset_sampler_rng(sampler._base_sampler)

        if isinstance(sampler, (NSGAIISampler, NSGAIIISampler)) and isinstance(
            sampler._child_generation_strategy, NSGAIIChildGenerationStrategy
        ):
            sampler._child_generation_strategy._rng.rng.seed(self._random_state)

        if isinstance(sampler, QMCSampler) and self._random_state is not None:
            sampler._seed = self._random_state



"""Module for fetching data from AWS S3 and storing it locally using the `S3Ingester` class."""
from __future__ import annotations

import hashlib
import os
from concurrent import futures
from dataclasses import dataclass
from datetime import date
from fnmatch import fnmatch
from pathlib import Path

import boto3
from boto3.s3.transfer import TransferConfig as BotoTransferConfig
from botocore.config import Config as BotoConfig
from tqdm.auto import tqdm

from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_ingester import BaseIngester, LocalFileEntry, LocalManifestHandler


logger = CustomLogger()
"""A module-level custom logger."""


@dataclass
class S3FileManifest:
    """Manifest entry for a single S3 file."""

    key: Path
    """Key of the file, the full path of the file in the S3 bucket including the key prefix."""

    size: int
    """Size of the file in bytes."""

    last_modified: date
    """Last modified date of the file."""


class S3Ingester(BaseIngester):
    """`S3Ingester` provides a convenient interface for fetching data from AWS S3 buckets and storing it locally.

    This class interacts with AWS S3 to download specified data from an S3 bucket.
    It supports manifest-based caching, enabling more efficient data fetching by verifying if the
    local dataset is up-to-date before downloading.
    """

    @auto_repr
    def __init__(
        self,
        s3_bucket_name: str,
        s3_key_prefix: str,
        file_pattern: str | list[str] = "*",
        dataset_id: str | None = None,
        destination_directory: str | Path = "data/ingest-s3",
        aws_profile_name: str | None = None,
        aws_region_name: str = "eu-west-1",
        num_workers: int = 64,
        check_s3_timestamps: bool = True,
    ) -> None:
        """Initializes the S3 bucket client, configures the destination directory, and sets client-related parameters.

        Note:
            The S3 bucket client is initialized using the provided AWS profile and region. If no profile is provided,
            the default profile will be used. If no region is provided, the default region will be used.

            The profile and region is read from the AWS credentials file located at '~/.aws/credentials'.

        Args:
            s3_bucket_name: Name of the S3 bucket containing the data.
            s3_key_prefix: Prefix of the S3 keys for the files to download
            file_pattern: Pattern to match the files to download, e.g. `*.csv` or [`*.csv`, `*.json`], etc.
                For more information, see https://docs.python.org/3/library/fnmatch.html.
            dataset_id: Id of the dataset to be used instead of the default fingerprint (MD5 hash of the bucket
                name, key prefix, and region name). Note that this will overwrite any existing dataset with the same
                name in the destination directory, so make sure to use a unique name.
            destination_directory: Directory to store the fetched data locally.
            aws_profile_name: AWS profile name to use.
            aws_region_name: AWS region name where the S3 bucket is located.
            num_workers: Number of workers to use for concurrent downloads.
            check_s3_timestamps: Whether to check if all S3 files have the same timestamp.

        Examples:
            >>> from mleko.dataset.sources import S3Ingester
            >>> s3_ingester = S3Ingester(
            ...     s3_bucket_name="mleko-datasets",
            ...     s3_key_prefix="kaggle/ashishpatel26/indian-food-101",
            ...     file_pattern="file_*.csv",
            ...     dataset_id="indian_food", # Optional, but will store the data in "./data/indian_food/" instead of
            ...                               # "./data/<fingerprint>/".
            ...     destination_directory="data",
            ...     aws_profile_name="mleko",
            ...     aws_region_name="eu-west-1",
            ...     num_workers=64,
            ...     check_s3_timestamps=True,
            ... )
            >>> s3_ingester.fetch_data()
            [PosixPath('data/indian_food/indian_food.csv')]
        """
        dataset_id = (
            dataset_id
            if dataset_id is not None
            else hashlib.md5((s3_bucket_name + s3_key_prefix + aws_region_name).encode()).hexdigest()
        )
        super().__init__(destination_directory, dataset_id)
        self._local_manifest_handler = LocalManifestHandler(
            self._destination_directory / f"{self._fingerprint}.manifest.json"
        )
        self._s3_bucket_name = s3_bucket_name
        self._s3_key_prefix = s3_key_prefix
        self._aws_profile_name = aws_profile_name
        self._aws_region_name = aws_region_name
        self._s3_client = self._get_s3_client(self._aws_profile_name, self._aws_region_name)
        self._num_workers = num_workers
        self._check_s3_timestamps = check_s3_timestamps

        if isinstance(file_pattern, str):
            file_pattern = [file_pattern]
        self._file_pattern = file_pattern

    def fetch_data(self, force_recompute: bool = False) -> list[Path]:
        """Downloads the data from the S3 bucket and stores it in the 'destination_directory'.

        If 'force_recompute' is False, verifies whether the data in the local 'destination_directory' is current
        with the S3 bucket contents based on the manifest file, and skips downloading if it is up to date.

        Args:
            force_recompute: Whether to force the data source to recompute its output, even if it already exists.

        Raises:
            Exception: If files in the S3 bucket have different last modified dates, indicating potential corruption
                       or duplication.

        Returns:
            A list of Path objects pointing to the downloaded data files.
        """
        self._s3_client = self._get_s3_client(self._aws_profile_name, self._aws_region_name)
        s3_manifest = self._build_s3_manifest()

        if self._check_s3_timestamps:
            modification_dates = {key.last_modified for key in s3_manifest}
            if len(modification_dates) > 1:
                error_msg = "Files in S3 are from muliples dates. This might mean the data is corrupted/duplicated."
                logger.error(error_msg)
                raise Exception(error_msg)

        if force_recompute:
            logger.info(
                f"\033[33mForce Cache Refresh\033[0m: Downloading files matching {self._file_pattern} from "
                f"{self._s3_bucket_name}/{self._s3_key_prefix} to {self._destination_directory} from S3."
            )
        else:
            if self._is_local_dataset_fresh(s3_manifest):
                logger.info(
                    "\033[32mCache Hit\033[0m: Local dataset is up to date with S3 bucket contents, "
                    "skipping download."
                )
                local_file_names = set(self._local_manifest_handler.get_file_names())
                s3_file_names: set[str] = {s3_file.key.name for s3_file in s3_manifest}
                files_to_delete = list(local_file_names.difference(s3_file_names))

                if len(files_to_delete) > 0:
                    logger.info(
                        f"Deleting {len(files_to_delete)} files from "
                        f"{self._destination_directory} that are no longer present in S3 or filtered out."
                    )

                self._delete_local_files(files_to_delete)
                self._local_manifest_handler.remove_files(files_to_delete)
                return self._get_full_file_paths(self._local_manifest_handler.get_file_names())

            logger.info(
                f"\033[31mCache Miss\033[0m: Downloading {self._s3_bucket_name}/{self._s3_key_prefix} to "
                f"{self._destination_directory} from S3."
            )

        self._delete_local_files(self._local_manifest_handler.get_file_names())
        keys_to_download: list[str] = [str(s3_file.key) for s3_file in s3_manifest]
        if len(keys_to_download) > 0:
            self._s3_fetch_all(keys_to_download)
            self._local_manifest_handler.set_files(
                [
                    LocalFileEntry(
                        name=Path(key).name, size=os.path.getsize(self._destination_directory / Path(key).name)
                    )
                    for key in keys_to_download
                ]
            )
            logger.info(f"Finished downloading {len(keys_to_download)} files from S3.")

        return self._get_full_file_paths(self._local_manifest_handler.get_file_names())

    def _build_s3_manifest(self) -> list[S3FileManifest]:
        """Builds a manifest from the S3 response.

        The manifest contains the S3 keys, sizes, and last modified dates of the files in the S3 bucket.

        Raises:
            FileNotFoundError: If no files matching the file pattern are found in the S3 bucket.

        Returns:
            A list of `S3FileManifest` objects containing the S3 keys, sizes, and last modified dates of the files in
            the S3 bucket.
        """
        resp = self._s3_client.list_objects(
            Bucket=self._s3_bucket_name,
            Prefix=self._s3_key_prefix,
        )

        s3_manifest: list[S3FileManifest] = [
            S3FileManifest(key=Path(key["Key"]), size=key["Size"], last_modified=key["LastModified"].date())
            for key in resp.get("Contents", [])
            if "LastModified" in key
            and "Key" in key
            and "Size" in key
            and any(fnmatch(Path(key["Key"]).name, pattern) for pattern in self._file_pattern)
        ]

        if len(s3_manifest) == 0:
            error_msg = (
                f"No files matching {self._file_pattern} found in S3 bucket "
                f"{self._s3_bucket_name}/{self._s3_key_prefix}."
            )
            logger.error(error_msg)
            raise FileNotFoundError(error_msg)
        logger.info(f"Found {len(s3_manifest)} file(s) matching any of {self._file_pattern} in S3 bucket.")

        return s3_manifest

    def _get_s3_client(
        self,
        aws_profile_name: str | None,
        aws_region_name: str,
    ):
        """Creates an S3 client using the provided AWS profile and region.

        Args:
            aws_profile_name: AWS profile name to use.
            aws_region_name: AWS region name where the S3 bucket is located.

        Returns:
            An S3 client configured with the specified profile and region.
        """
        credentials = boto3.Session(
            region_name=aws_region_name,
            profile_name=aws_profile_name,
        ).get_credentials()
        client_config = BotoConfig(max_pool_connections=100)
        return boto3.client(
            "s3",  # type: ignore
            aws_access_key_id=credentials.access_key,
            aws_secret_access_key=credentials.secret_key,
            aws_session_token=credentials.token,
            region_name=aws_region_name,
            config=client_config,
        )

    def _s3_fetch_file(self, key: str) -> None:
        """Downloads a single file from the S3 bucket to the destination specified in the constructor.

        Args:
            key: S3 key of the file to download.
        """
        gb = 1024**3
        transfer_config = BotoTransferConfig(
            use_threads=False,
            multipart_threshold=int(0.5 * gb),  # Multipart transfer if file > 500MB
        )
        file_path = self._destination_directory / Path(key).name
        with open(file_path, "wb") as data:
            self._s3_client.download_fileobj(
                Bucket=self._s3_bucket_name,
                Key=key,
                Fileobj=data,
                Config=transfer_config,
            )

    def _s3_fetch_all(self, keys: list[str]) -> None:
        """Downloads all specified files from the S3 bucket to the local directory concurrently.

        Args:
            keys: List of S3 keys for the files to download.
        """
        with tqdm(total=len(keys), desc="Downloading files from S3") as pbar:
            with futures.ThreadPoolExecutor(max_workers=self._num_workers) as executor:
                for _ in executor.map(
                    self._s3_fetch_file,
                    keys,
                ):
                    pbar.update(1)

    def _is_local_dataset_fresh(self, s3_manifest: list[S3FileManifest]) -> bool:
        """Checks if the local dataset is up-to-date with the S3 manifest.

        Args:
            s3_manifest: Manifest built from S3 response.

        Returns:
            True if the local dataset is up-to-date, False otherwise.
        """
        for s3_file in s3_manifest:
            local_file_path = self._destination_directory / s3_file.key.name
            if not local_file_path.exists() or s3_file.size != os.path.getsize(local_file_path):
                return False
        return True



"""Module for the base hyperparameter tuning class."""
from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Hashable

import vaex

from mleko.cache.fingerprinters.vaex_fingerprinter import VaexFingerprinter
from mleko.cache.handlers.joblib_cache_handler import JOBLIB_CACHE_HANDLER
from mleko.cache.handlers.pickle_cache_handler import PICKLE_CACHE_HANDLER
from mleko.cache.lru_cache_mixin import LRUCacheMixin
from mleko.dataset.data_schema import DataSchema
from mleko.model.base_model import HyperparametersType


class BaseTuner(LRUCacheMixin, ABC):
    """Abstract base class for hyperparameter tuners."""

    def __init__(
        self,
        cache_directory: str | Path,
        cache_size: int,
    ) -> None:
        """Initializes the `BaseTuner` with an output directory.

        Args:
            cache_directory: The target directory where the output is to be saved.
            cache_size: The maximum number of cache entries.
        """
        super().__init__(cache_directory, cache_size)

    def tune(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
    ) -> tuple[HyperparametersType, float | list[float] | tuple[float, ...], Any]:
        """Perform the hyperparameter tuning on the given DataFrame.

        Args:
            data_schema: Data schema for the DataFrame.
            dataframe: DataFrame to be tuned on.
            cache_group: The cache group to use for caching.
            force_recompute: Weather to force recompute the result.

        Returns:
            Tuple containing the best hyperparameters, the best score, and a dictionary
            containing any additional information about the tuning process. The dictionary
            is specific to each tuner, please refer to the documentation of the tuner
            for more information.
        """
        return self._cached_execute(
            lambda_func=lambda: self._tune(data_schema, dataframe),
            cache_key_inputs=[self._fingerprint(), str(data_schema), (dataframe, VaexFingerprinter())],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=[JOBLIB_CACHE_HANDLER, JOBLIB_CACHE_HANDLER, PICKLE_CACHE_HANDLER],
        )

    @abstractmethod
    def _tune(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame
    ) -> tuple[HyperparametersType, float | list[float] | tuple[float, ...], Any]:
        """Perform the hyperparameter tuning.

        Args:
            data_schema: Data schema for the DataFrame.
            dataframe: DataFrame to be tuned on.

        Raises:
            NotImplementedError: Must be implemented by subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def _fingerprint(self) -> Hashable:
        """Returns a hashable object that uniquely identifies the hyperparameter tuning process.

        The base implementation fingerprints the class name of the tuner.

        Note:
            Subclasses should call the parent method and include the result in the hashable object along with any
            other parameters that uniquely identify the tuner. All attributes that are used in the
            tuner that affect the result of the fitting and transforming should be included in the hashable object.

        Returns:
            Hashable: A hashable object that uniquely identifies the hyperparameter tuning process.
        """
        return self.__class__.__name__



"""The module containing the base class for data splitter."""
from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path

import vaex

from mleko.cache.lru_cache_mixin import LRUCacheMixin


class BaseSplitter(LRUCacheMixin, ABC):
    """Abstract base class for data splitter.

    Will cache the split dataframes in the output directory.
    """

    def __init__(self, cache_directory: str | Path, cache_size: int):
        """Initializes the `BaseSplitter` with an output directory.

        Args:
            cache_directory: The target directory where the split dataframes are to be saved.
            cache_size: The maximum number of cache entries.
        """
        super().__init__(cache_directory, cache_size)

    @abstractmethod
    def split(
        self, dataframe: vaex.DataFrame, cache_group: str | None = None, force_recompute: bool = False
    ) -> tuple[vaex.DataFrame, vaex.DataFrame]:
        """Abstract method to split the given dataframe into two parts.

        Args:
            dataframe: The dataframe to be split.
            cache_group: The cache group to use.
            force_recompute: Forces recomputation if True, otherwise reads from the cache if available.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError



"""Subpackage containing specialized pipeline steps for various data processing tasks.

This subpackage offers a collection of pipeline steps, each designed for a specific purpose: data ingestion,
data conversion, and other data manipulation tasks. By using these unique steps sequentially, you can create a
complete data processing workflow within the pipeline.
"""
from __future__ import annotations

from .convert_step import ConvertStep
from .feature_select_step import FeatureSelectStep
from .ingest_step import IngestStep
from .model_step import ModelStep
from .split_step import SplitStep
from .transform_step import TransformStep
from .tune_step import TuneStep


__all__ = ["IngestStep", "ConvertStep", "SplitStep", "FeatureSelectStep", "TransformStep", "ModelStep", "TuneStep"]



"""Module for the label encoder transformer."""
from __future__ import annotations

from pathlib import Path
from typing import Hashable

import vaex
import vaex.ml

from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr

from .base_transformer import BaseTransformer


logger = CustomLogger()
"""A module-level logger for the module."""


class LabelEncoderTransformer(BaseTransformer):
    """Transforms features using label encoding."""

    @auto_repr
    def __init__(
        self,
        cache_directory: str | Path,
        features: list[str] | tuple[str, ...],
        allow_unseen: bool = True,
        cache_size: int = 1,
    ) -> None:
        """Initializes the transformer.

        Uses the `vaex.ml.LabelEncoder` transformer, which encodes categorical features with integer values between 0
        and n_classes-1. If a value is not seen during fitting, it will be encoded as -1, unless `allow_unseen` is
        set to False, in which case an error will be raised.

        Warning:
            Should only be used with categorical features.

        Args:
            cache_directory: Directory where the cache will be stored locally.
            features: List of feature names to be used by the transformer.
            allow_unseen: Whether to allow unseen values once the transformer is fitted.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.dataset.transform import LabelEncoderTransformer
            >>> df = vaex.from_arrays(
            ...     a=["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
            ...     b=["a", "a", "a", "a", None, None, None, None, None, None],
            ...     c=["a", "b", "b", "b", "b", "b", None, None, None, None],
            ... )
            >>> ds = DataSchema(
            ...     categorical=["a", "b", "c"],
            ... )
            >>> _, _, df = LabelEncoderTransformer(
            ...     cache_directory=".",
            ...     features=["a", "b"],
            ...     allow_unseen=True,
            ... ).fit_transform(ds, df)
            >>> df["a"].tolist()
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            >>> df["b"].tolist()
            [1, 1, 1, 1, 0, 0, 0, 0, 0, 0]
        """
        super().__init__(cache_directory, features, cache_size)
        self._allow_unseen = allow_unseen
        self._transformer = vaex.ml.LabelEncoder(allow_unseen=self._allow_unseen, features=self._features, prefix="")

    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.ml.LabelEncoder]:
        """Fits the transformer on the given DataFrame.

        Args:
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to fit the transformer on.

        Returns:
            Updated data schema and fitted transformer.
        """
        logger.info(f"Fitting label encoder transformer ({len(self._features)}): {self._features}.")
        self._transformer.fit(dataframe)

        return data_schema, self._transformer

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Transforms the features of the given DataFrame using label encoding.

        Args:
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to transform.

        Returns:
            Updated data schema and transformed DataFrame.
        """
        logger.info(f"Transforming features using label encoding ({len(self._features)}): {self._features}.")
        transformed_df = self._transformer.transform(dataframe)

        return data_schema, transformed_df

    def _fingerprint(self) -> Hashable:
        """Returns the fingerprint of the transformer.

        Appends the `allow_unseen` parameter to the fingerprint of the base class.

        Returns:
            A hashable object that uniquely identifies the transformer.
        """
        return super()._fingerprint(), self._allow_unseen



"""This module contains the CacheHandler for reading a writing `vaex` DataFrames to disk."""
import warnings
from pathlib import Path

import vaex
from tqdm.auto import tqdm

from mleko.utils.tqdm_helpers import set_tqdm_percent_wrapper

from .base_cache_handler import CacheHandler


def read_vaex_dataframe(cache_file_path: Path) -> vaex.DataFrame:
    """Reads a cache file containing a `vaex` DataFrame.

    Args:
        cache_file_path: The path of the cache file to be read.

    Returns:
        The contents of the cache file as a DataFrame.
    """
    return vaex.open(cache_file_path)


def write_vaex_dataframe(cache_file_path: Path, output: vaex.DataFrame) -> None:
    """Writes the results of the DataFrame conversion to a file.

    Args:
        cache_file_path: The path of the cache file to be written.
        output: The Vaex DataFrame to be saved in the cache file.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "invalid value encountered in cast")
        with tqdm(total=100, desc=f"Writing DataFrame to {cache_file_path.suffix} file") as pbar:
            output.export(
                cache_file_path,
                progress=set_tqdm_percent_wrapper(pbar),
                parallel=True,
                chunk_size=100_000,
            )


VAEX_DATAFRAME_CACHE_HANDLER = CacheHandler(writer=write_vaex_dataframe, reader=read_vaex_dataframe, suffix="hdf5")
"""A CacheHandler for `vaex` DataFrames."""



"""Module containing the ModelStep class."""
from __future__ import annotations

from typing import Literal

from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.model.base_model import BaseModel
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class ModelStep(PipelineStep):
    """Pipeline step for model training and prediction."""

    _num_inputs = 4
    """Number of inputs expected by the ModelStep."""

    _num_outputs = 1
    """Number of outputs expected by the ModelStep."""

    @auto_repr
    def __init__(
        self,
        model: BaseModel,
        action: Literal["fit", "transform", "fit_transform"],
        inputs: list[str] | tuple[str, ...] | tuple[()] = (),
        outputs: list[str] | tuple[str, ...] | tuple[()] = (),
        cache_group: str | None = None,
    ) -> None:
        """Initialize the ModelStep with the specified model.

        The action parameter specifies whether the model should be fitted, transformed, or both. If the action
        is "fit" the `ModelStep` will return the fitted model. If the action is "transform" the
        `ModelStep` will return the predictions of the DataFrame. If the action is "fit_transform" the `ModelStep`
        will return both the fitted model and the predictions of the DataFrame in the specified order.

        Note:
            The hyperparameters input is optional. If it is not provided, the model will be trained using the
            default hyperparameters configured in the model.

        Args:
            model: The model used for training and prediction.
            action: The action to perform, one of "fit", "transform", or "fit_transform".
            inputs: List or tuple of input keys expected by this step. If the action is "fit" or "fit_transform",
                should contain four keys, corresponding to the DataSchema, the training DataFrame to be fitted,
                the validation DataFrame to be used for validation, and the hyperparameters to be used for training.
                If the action is "transform", should contain two keys, corresponding to the DataSchema, and the
                DataFrame to be transformed.
            outputs: List or tuple of output keys produced by this step. If the action is "fit", should contain two
                keys, corresponding to the fitted model and the metrics dictionary. If the action is "transform",
                should contain a single key, corresponding to the transformed DataFrame. If the action is
                "fit_transform", should contain four keys, corresponding to the fitted model, the metrics
                dictionary, the transformed training DataFrame, and the transformed validation DataFrame.
            cache_group: The cache group to use.

        Raises:
            ValueError: If action is not one of "fit", "transform", or "fit_transform".
        """
        if action not in ("fit", "transform", "fit_transform"):
            raise ValueError(
                f"Invalid action: {action}. Expected one of 'fit', 'transform', or 'fit_transform'."
            )  # pragma: no cover

        if action == "fit":
            self._num_outputs = 2
        if action == "transform":
            self._num_inputs = 2
        if action == "fit_transform":
            self._num_outputs = 4

        super().__init__(inputs, outputs, cache_group)
        self._model = model
        self._action = action

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform actions using the configured model.

        Args:
            data_container: Contains the input.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            ValueError: If data container contains invalid data - not a vaex DataFrame.

        Returns:
            A DataContainer containing the output of the action performed by the step, either the fitted model,
            the predictions on the DataFrame, or both.
        """
        data_schema = data_container.data[self._inputs[0]]
        if not isinstance(data_schema, DataSchema):
            raise ValueError(f"Invalid data type: {type(data_schema)}. Expected DataSchema.")

        dataframe = data_container.data[self._inputs[1]]
        if not isinstance(dataframe, DataFrame):
            raise ValueError(f"Invalid data type: {type(dataframe)}. Expected vaex DataFrame.")

        df_validation = hyperparameters = None
        if self._action != "transform":
            df_validation = data_container.data[self._inputs[2]]
            if not isinstance(df_validation, DataFrame):
                raise ValueError(f"Invalid data type: {type(df_validation)}. Expected vaex DataFrame.")

            hyperparameters = {}
            if len(self._inputs) == 4:
                hyperparameters = data_container.data.get(self._inputs[3])
            if not isinstance(hyperparameters, dict):
                raise ValueError(f"Invalid data type: {type(hyperparameters)}. Expected HyperparametersType.")

        if self._action == "fit" and df_validation is not None:
            model, metrics = self._model.fit(
                data_schema, dataframe, df_validation, hyperparameters, self._cache_group, force_recompute
            )
            data_container.data[self._outputs[0]] = model
            data_container.data[self._outputs[1]] = metrics
        elif self._action == "transform":
            df = self._model.transform(data_schema, dataframe, self._cache_group, force_recompute)
            data_container.data[self._outputs[0]] = df
        elif self._action == "fit_transform" and df_validation is not None:
            model, metrics, df, df_validation = self._model.fit_transform(
                data_schema, dataframe, df_validation, hyperparameters, self._cache_group, force_recompute
            )
            data_container.data[self._outputs[0]] = model
            data_container.data[self._outputs[1]] = metrics
            data_container.data[self._outputs[2]] = df
            data_container.data[self._outputs[3]] = df_validation
        return data_container

    def _validate_inputs(self) -> None:
        """Check the input keys for compliance with this step's requirements.

        Raises:
            ValueError: If the PipelineStep has an invalid number of inputs.
        """
        if len(self._inputs) != self._num_inputs and len(self._inputs) != self._num_inputs - 1:
            raise ValueError(f"{self.__class__.__name__} must have exactly {self._num_inputs} input(s).")



"""The subpackage contains the core data processing functionality of the `MLEKO` library.

This subpackage focuses on handling different aspects of data processing, including fetching data from various
sources, converting between file formats, filtering, and feature engineering.

The following submodules are provided:
    - `ingest`: The submodule provides functionality for fetching data from various sources.
    - `convert`: The submodule provides functionality for converting between different file formats.
    - `split`: The submodule provides functionality for splitting data into multiple parts.
    - `feature_select`: The submodule provides functionality for selecting features from data.
"""
from __future__ import annotations

from .data_schema import DataSchema


__all__ = ["DataSchema"]



"""Module for the base transformer class."""
from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Hashable

import vaex

from mleko.cache.fingerprinters.vaex_fingerprinter import VaexFingerprinter
from mleko.cache.handlers.joblib_cache_handler import JOBLIB_CACHE_HANDLER
from mleko.cache.handlers.vaex_cache_handler import VAEX_DATAFRAME_CACHE_HANDLER
from mleko.cache.lru_cache_mixin import LRUCacheMixin
from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger


logger = CustomLogger()
"""The logger for the module."""


class BaseTransformer(LRUCacheMixin, ABC):
    """Abstract class for feature transformation.

    The feature transformation process is implemented in the `fit`, `transform`, and `fit_transform` methods. The
    `fit` method fits the transformer to the specified DataFrame, the `transform` method transforms the specified
    features in the DataFrame, and the `fit_transform` method fits the transformer to the specified DataFrame and
    transforms the specified features in the DataFrame.
    """

    def __init__(
        self,
        cache_directory: str | Path,
        features: list[str] | tuple[str, ...],
        cache_size: int,
    ) -> None:
        """Initializes the transformer and ensures the destination directory exists.

        Args:
            cache_directory: Directory where the cache will be stored locally.
            features: List of feature names to be used by the transformer.
            cache_size: The maximum number of cache entries to keep in the cache.
        """
        super().__init__(cache_directory, cache_size)
        self._features: tuple[str, ...] = tuple(features)
        self._transformer = None

    def fit(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
    ) -> tuple[DataSchema, Any]:
        """Fits the transformer to the specified DataFrame, using the cached result if available.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame to be fitted.
            cache_group: The cache group to use.
            force_recompute: Whether to force the fitting to be recomputed even if the result is cached.

        Returns:
            Updated data schema and fitted transformer.
        """
        ds, transformer = self._cached_execute(
            lambda_func=lambda: self._fit(data_schema, dataframe),
            cache_key_inputs=[self._fingerprint(), str(data_schema), (dataframe, VaexFingerprinter())],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=JOBLIB_CACHE_HANDLER,
        )
        self._assign_transformer(transformer)
        return ds, transformer

    def transform(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
    ) -> tuple[DataSchema, vaex.DataFrame]:
        """Transforms the specified features in the DataFrame, using the cached result if available.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame to be transformed.
            cache_group: The cache group to use.
            force_recompute: Whether to force the transformation to be recomputed even if the result is cached.

        Raises:
            RuntimeError: If the transformer has not been fitted.

        Returns:
            Updated data schema and transformed DataFrame.
        """
        if self._transformer is None:
            raise RuntimeError("Transformer must be fitted before it can be used to transform data.")

        ds, df = self._cached_execute(
            lambda_func=lambda: self._transform(data_schema, dataframe),
            cache_key_inputs=[self._fingerprint(), str(data_schema), (dataframe, VaexFingerprinter())],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=[JOBLIB_CACHE_HANDLER, VAEX_DATAFRAME_CACHE_HANDLER],
        )
        return ds, df

    def fit_transform(
        self,
        data_schema: DataSchema,
        dataframe: vaex.DataFrame,
        cache_group: str | None = None,
        force_recompute: bool = False,
    ) -> tuple[DataSchema, Any, vaex.DataFrame]:
        """Fits the transformer to the specified DataFrame and transforms the specified features in the DataFrame.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame used for fitting and transformation.
            cache_group: The cache group to use.
            force_recompute: Whether to force the fitting and transformation to be recomputed even if the result is

        Returns:
            Tuple of updated data schema, fitted transformer, and transformed DataFrame.
        """
        ds, transformer, df = self._cached_execute(
            lambda_func=lambda: self._fit_transform(data_schema, dataframe),
            cache_key_inputs=[self._fingerprint(), str(data_schema), (dataframe, VaexFingerprinter())],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=[JOBLIB_CACHE_HANDLER, JOBLIB_CACHE_HANDLER, VAEX_DATAFRAME_CACHE_HANDLER],
        )
        self._assign_transformer(transformer)
        return ds, transformer, df

    def _fit_transform(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame
    ) -> tuple[DataSchema, Any, vaex.DataFrame]:
        """Fits the transformer to the specified DataFrame and transforms the specified features in the DataFrame.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame used for fitting and transformation.

        Returns:
            Tuple of updated data schema, fitted transformer, and transformed DataFrame.
        """
        ds, transformer = self._fit(data_schema, dataframe)
        ds, df = self._transform(data_schema, dataframe)
        return ds, transformer, df

    def _assign_transformer(self, transformer: Any) -> None:
        """Assigns the specified transformer to the transformer attribute.

        Can be overridden by subclasses to assign the transformer using a different method.

        Args:
            transformer: Transformer to be assigned.
        """
        self._transformer = transformer

    @abstractmethod
    def _fit(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, Any]:
        """Fits the transformer to the specified DataFrame.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame to be fitted.

        Raises:
            NotImplementedError: Must be implemented by subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Transforms the specified features in the DataFrame.

        Args:
            data_schema: Data schema of the DataFrame.
            dataframe: DataFrame to be transformed.

        Raises:
            NotImplementedError: Must be implemented by subclasses.
        """
        raise NotImplementedError

    @abstractmethod
    def _fingerprint(self) -> Hashable:
        """Returns a hashable object that uniquely identifies the transformer.

        The base implementation fingerprints the class name and the features used by the transformer.

        Note:
            Subclasses should call the parent method and include the result in the hashable object along with any
            other parameters that uniquely identify the transformer. All attributes that are used in the
            transformer that affect the result of the transformation should be included in the hashable object.

        Returns:
            Hashable object that uniquely identifies the transformer.
        """
        return self.__class__.__name__, self._features



"""The module provides a `RandomSplitter` class for splitting `vaex` DataFrames randomly.

The splitter can be used to split a `vaex` DataFrame into two parts, with the split being performed randomly. The split
can be stratified by specifying a column name to use for stratification.
"""
from __future__ import annotations

from pathlib import Path

import vaex
from sklearn.model_selection import train_test_split

from mleko.cache.fingerprinters import VaexFingerprinter
from mleko.cache.handlers.vaex_cache_handler import VAEX_DATAFRAME_CACHE_HANDLER
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr
from mleko.utils.vaex_helpers import get_column, get_filtered_df

from .base_splitter import BaseSplitter


logger = CustomLogger()
"""A CustomLogger instance that's used throughout the module for logging."""


class RandomSplitter(BaseSplitter):
    """A class that handles random splitting of `vaex` DataFrames.

    This class provides a method for splitting a `vaex` DataFrame into two parts, with the split being performed
    randomly. The split can be stratified by specifying a column name to use for stratification.
    """

    @auto_repr
    def __init__(
        self,
        cache_directory: str | Path,
        data_split: tuple[float, float] = (0.80, 0.20),
        shuffle: bool = True,
        stratify: str | None = None,
        random_state: int | None = None,
        cache_size: int = 1,
    ):
        """Initializes the `RandomSplitter` with the given parameters.

        Note:
            The stratification is performed before the split, meaning that the split will be performed on the stratified
            data. For example, if the data is split into 80% train and 20% test, and the stratification column contains
            80% of the rows with value 0 and 20% of the rows with value 1, the resulting split will contain 80% of the
            rows with value 0 and 20% of the rows with value 1.

        Warning:
            If `stratify` is not None and the target column is a string/categorical, the target feature must not have
            any missing values. Please make sure to handle missing values before using this splitter by either dropping
            the rows with missing values, imputing the missing values, or transforming the target to numeric or boolean.

        Args:
            cache_directory: The target directory where the split dataframes are to be saved.
            data_split: A tuple containing the desired split percentages or weights for the train and test dataframes.
                If the sum of the values is not equal to 1, the values will be normalized. Meaning, if the values are
                (0.90, 0.20), the resulting split will be (0.818, 0.182).
            shuffle: Whether to shuffle the data before splitting.
            stratify: The name of the column to use for stratification. If None, stratification will not be performed.
            random_state: The seed to use for random number generation.
            cache_size: The maximum number of entries to keep in the cache.

        Example:
            >>> import vaex
            >>> from mleko.data.split import RandomSplitter
            >>> df = vaex.from_arrays(x=[1, 2, 3, 4], y=[0, 1, 1, 0])
            >>> splitter = RandomSplitter(cache_directory="cache", data_split=(0.50, 0.50), shuffle=True, stratify="y")
            >>> df_train, df_test = splitter.split(df)
            >>> df_train
                #    x    y
                0    1    0
                1    3    1
            >>> df_test
                #    x    y
                0    2    1
                1    4    0
        """
        super().__init__(cache_directory, cache_size)
        self._idx2_size = [split / sum(data_split) for split in data_split][1]
        self._shuffle = shuffle
        self._stratify = stratify
        self._random_state = random_state

    def split(
        self, dataframe: vaex.DataFrame, cache_group: str | None = None, force_recompute: bool = False
    ) -> tuple[vaex.DataFrame, vaex.DataFrame]:
        """Split the given dataframe into two parts.

        Splits the dataframe into train and test sets according to the proportions, shuffle,
        stratification, and random state provided during initializing the splitter. Will read from the cache if
        available, unless `force_recompute=True`.

        Args:
            dataframe: The dataframe to be split.
            cache_group: The cache group to use.
            force_recompute: Whether to force recompute the split, even if the cache is available.

        Returns:
            A tuple containing the split dataframes.
        """
        return self._cached_execute(
            lambda_func=lambda: self._split(dataframe),
            cache_key_inputs=[
                self._idx2_size,
                self._shuffle,
                self._stratify,
                self._random_state,
                (dataframe, VaexFingerprinter()),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=VAEX_DATAFRAME_CACHE_HANDLER,
        )

    def _split(self, dataframe: vaex.DataFrame) -> tuple[vaex.DataFrame, vaex.DataFrame]:
        """Split the given dataframe into two parts.

        Args:
            dataframe: The dataframe to be split.

        Returns:
            A tuple containing the split dataframes.
        """
        index_name = "index"
        df = dataframe.copy()
        df[index_name] = vaex.vrange(0, df.shape[0])
        index = get_column(df, index_name)
        target = get_column(df, self._stratify).to_numpy() if self._stratify else None

        if self._shuffle:
            logger.info("Shuffling data before splitting.")
        if target is None:
            logger.info("Splitting data without stratification.")
        else:
            logger.info(f"Splitting data with stratification on column {self._stratify!r}.")

        idx1, idx2 = train_test_split(
            index.values,
            test_size=self._idx2_size,
            random_state=self._random_state,
            shuffle=self._shuffle,
            stratify=target,
        )

        df1 = get_filtered_df(df, index.isin(idx1)).extract()
        df2 = get_filtered_df(df, index.isin(idx2)).extract()
        logger.info(f"Split dataframe into two dataframes with shapes {df1.shape} and {df2.shape}.")
        df1.delete_virtual_column(index_name)
        df2.delete_virtual_column(index_name)
        return df1, df2



"""Module for the LightGBM model."""
from __future__ import annotations

from pathlib import Path
from typing import Any, Hashable, Literal

import lightgbm as lgb
import vaex

from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr
from mleko.utils.vaex_helpers import get_column, get_columns

from .base_model import BaseModel, HyperparametersType


logger = CustomLogger()
"""The logger for the module."""

LGBMObjectiveType = Literal[
    "regression",
    "regression_l1",
    "huber",
    "fair",
    "poisson",
    "quantile",
    "mape",
    "gamma",
    "tweedie",
    "binary",
    "multiclass",
    "multiclassova",
    "cross_entropy",
    "cross_entropy_lambda",
    "lambdarank",
    "rank_xendcg",
]
"""The type of the objective function."""

LGBMBoostingType = Literal["gbdt", "rf", "dart"]
"""The type of the boosting algorithm."""

LGBMDataSampleStrategyType = Literal["bagging", "goss"]
"""The type of the data sampling strategy."""

LGBMTreeLearnerType = Literal["serial", "feature", "data", "voting"]
"""The type of the tree learner."""

LGBMMetricType = Literal[
    "",
    "None",
    "l1",
    "l2",
    "rmse",
    "quantile",
    "mape",
    "huber",
    "fair",
    "poisson",
    "gamma",
    "gamma_negative",
    "tweedie",
    "ndcg",
    "map",
    "auc",
    "average_precision",
    "binary_logloss",
    "binary_error",
    "auc_mu",
    "multi_logloss",
    "multi_error",
    "cross_entropy",
    "cross_entropy_lambda",
    "kullback_leibler",
]
"""The type of the metric."""

LGBMMonotoneConstraintsMethodType = Literal["basic", "intermediate", "advanced"]
"""The type of the monotone constraints method."""


class LGBMModel(BaseModel):
    """Wrapper for the LightGBM model.

    Full documentation of the LightGBM model can be found here https://lightgbm.readthedocs.io/en/latest/index.html.
    """

    @auto_repr
    def __init__(
        self,
        cache_directory: str | Path,
        target: str,
        features: list[str] | tuple[str, ...] | None = None,
        ignore_features: list[str] | tuple[str, ...] | None = None,
        objective: LGBMObjectiveType = "regression",
        boosting: LGBMBoostingType = "gbdt",
        data_sample_strategy: LGBMDataSampleStrategyType = "bagging",
        num_iterations: int = 100,
        learning_rate: float = 0.1,
        num_leaves: int = 31,
        tree_learner: LGBMTreeLearnerType = "serial",
        random_state: int | None = None,
        force_col_wise: bool = False,
        force_row_wise: bool = False,
        histogram_pool_size: float = -1,
        max_depth: int = -1,
        min_data_in_leaf: int = 20,
        min_sum_hessian_in_leaf: float = 1e-3,
        bagging_fraction: float = 1.0,
        pos_bagging_fraction: float = 1.0,
        neg_bagging_fraction: float = 1.0,
        bagging_freq: int = 0,
        feature_fraction: float = 1.0,
        feature_fraction_bynode: float = 1.0,
        extra_trees: bool = False,
        early_stopping_round: int | None = None,
        first_metric_only: bool = False,
        max_delta_step: float = 0.0,
        lambda_l1: float = 0.0,
        lambda_l2: float = 0.0,
        linear_lambda: float = 0.0,
        min_gain_to_split: float = 0.0,
        drop_rate: float = 0.1,
        max_drop: int = 50,
        skip_drop: float = 0.5,
        xgboost_dart_mode: bool = False,
        uniform_drop: bool = False,
        top_rate: float = 0.2,
        other_rate: float = 0.1,
        min_data_per_group: int = 100,
        max_cat_threshold: int = 32,
        cat_l2: float = 10.0,
        cat_smooth: float = 10.0,
        max_cat_to_onehot: int = 4,
        top_k: int = 20,
        monotone_constraints: list[int] | tuple[int, ...] | None = None,
        monotone_constraints_method: LGBMMonotoneConstraintsMethodType = "basic",
        monotone_penalty: float = 0.0,
        feature_contri: list[float] | tuple[float, ...] | None = None,
        path_smooth: float = 0.0,
        verbosity: int = 1,
        log_evaluation_period: int | None = 10,
        use_quantized_grad: bool = False,
        num_grad_quant_bins: int = 4,
        quant_train_renew_leaf: bool = False,
        stochastic_rounding: bool = True,
        linear_tree: bool = False,
        max_bin: int = 255,
        max_bin_by_feature: list[int] | tuple[int, ...] | None = None,
        min_data_in_bin: int = 3,
        bin_construct_sample_cnt: int = 200000,
        is_enable_sparse: bool = True,
        enable_bundle: bool = True,
        use_missing: bool = True,
        zero_as_missing: bool = False,
        feature_pre_filter: bool = True,
        num_class: int = 1,
        is_unbalance: bool = False,
        scale_pos_weight: float = 1.0,
        sigmoid: float = 1.0,
        boost_from_average: bool = True,
        reg_sqrt: bool = False,
        alpha: float = 0.9,
        fair_c: float = 1.0,
        poisson_max_delta_step: float = 0.7,
        tweedie_variance_power: float = 1.5,
        lambdarank_truncation_level: int = 30,
        lambdarank_norm: bool = True,
        metric: LGBMMetricType | list[LGBMMetricType] = "",
        metric_freq: int = 1,
        is_provide_training_metric: bool = False,
        eval_at: list[int] | tuple[int, ...] | None = None,
        multi_error_top_k: int = 1,
        auc_mu_weights: list[float] | tuple[float, ...] | None = None,
        cache_size: int = 1,
    ) -> None:
        """Initialize the LightGBM model with the given hyperparameters.

        Note:
            Features and ignore_features are mutually exclusive. If both are provided, a `ValueError` will be raised.
            By default, all features are used. If ignore_features is provided, all features except the ones in
            ignore_features will be used. If features is provided, only the features in features will be used.


        Args:
            cache_directory: The target directory where the model will be saved.
            target: The name of the target feature.
            features: The names of the features to be used as input for the model.
            ignore_features: The names of the features to be ignored.
            objective: The objective function to be used.
            boosting: The type of boosting algorithm.
            data_sample_strategy: How data sampling should be performed.
            num_iterations: The number of iterations to perform.
            learning_rate: The learning rate.
            num_leaves: The maximum number of leaves in one tree.
            tree_learner: The type of tree learner.
            random_state: The seed to use for random number generation.
            force_col_wise: Force col-wise histogram building.
            force_row_wise: Force row-wise histogram building.
            histogram_pool_size: Size of pooled memory for histograms in MB.
            max_depth: The maximum depth of a tree.
            min_data_in_leaf: The minimum number of data in one leaf.
            min_sum_hessian_in_leaf: The minimum sum of hessian in one leaf.
            bagging_fraction: The fraction of data to be used when bagging.
            pos_bagging_fraction: The fraction of positive data to be used when bagging,
                only used in binary application.
            neg_bagging_fraction: The fraction of negative data to be used when bagging,
                only used in binary application.
            bagging_freq: Frequency for bagging, 0 means disable bagging; k means perform bagging at every k iteration.
                At every bagging iteration, LightGBM will randomly select `baggig_fraction` * `data_size` data from
                training data to train for the next `k` iterations.
            feature_fraction: The fraction of features to be used when building each tree.
            feature_fraction_bynode: The fraction of features to be used for each tree node.
            extra_trees: Whether to use extremely randomized trees.
            early_stopping_round: Will stop training if one metric of one validation data doesn't improve in last
                `early_stopping_round` rounds.
            first_metric_only: Whether to only use the first metric for early stopping.
            max_delta_step: The maximum delta step for tree leaf output.
            lambda_l1: L1 regularization.
            lambda_l2: L2 regularization.
            linear_lambda: Linear tree regularization.
            min_gain_to_split: The minimal gain to perform split.
            drop_rate: Dropout rate, only used in `dart` boosting, will drop trees with probability `drop_rate`.
            max_drop: Maximum number of dropped trees during one iteration, only used in `dart` boosting.
            skip_drop: Probability of skipping the dropout procedure during one iteration, only
                used in `dart` boosting.
            xgboost_dart_mode: Whether to use xgboost `dart` boosting.
            uniform_drop: Whether to use uniform drop, only used in `dart` boosting.
            top_rate: The retain ratio of large gradient data, only used in `goss` boosting.
            other_rate: The retain ratio of small gradient data, only used in `goss` boosting.
            min_data_per_group: The minimum number of data per categorical group.
            max_cat_threshold: Limit number of split points for categorical features.
            cat_l2: L2 regularization in categorical split.
            cat_smooth: Can reduce the effect of noises in categorical features, especially for
                categories with few data.
            max_cat_to_onehot: When number of categories of one feature smaller than or equal to `max_cat_to_onehot`,
                one-vs-other split algorithm will be used.
            top_k: Set this to larger value for more accurate result, but it will slow down the training speed.
            monotone_constraints: The constraints for monotonicity, set it to a list with the same size as the number
                of features to turn on the monotonicity constraint. `1` means increasing constraint, `-1` means
                decreasing constraint, and `0` means no constraint.
            monotone_constraints_method: The method to use to deal with monotone constraints, only used when
                `monotone_constraints` is not None.
            monotone_penalty: The penalty for monotone constraints.
            feature_contri: Used to control feature's split gain, will
                use gain[i] = max(0, feature_contri[i]) * gain[i] to replace the original gain of the ith feature.
            path_smooth: Controls the smoothing applied to tree nodes.
            verbosity: Controls the level of LightGBM's verbosity.
            log_evaluation_period: The period to log the evaluation results, if `None`, results will not be logged.
            use_quantized_grad: Whether to use gradiant quantization when training.
            num_grad_quant_bins: Number of bin to quantization gradients and hessians.
            quant_train_renew_leaf: Whether to renew the leaf values with original gradients when quantized training.
            stochastic_rounding: Whether to use stochastic rounding in gradient quantization.
            linear_tree: Fit piecewise linear gradient boosting tree.
            max_bin: Max number of bins that feature values will be bucketed in.
            max_bin_by_feature: Max number of bins that feature values will be bucketed in, per feature.
            min_data_in_bin: Min number of data inside one bin.
            bin_construct_sample_cnt: Number of data that sampled to construct histogram bins.
            is_enable_sparse: Used to enable/disable the sparse optimization.
            enable_bundle: Set this to `False` to disable Exclusive Feature Bundling (EFB).
            use_missing: Set this to `False` to disable the special handle of missing value.
            zero_as_missing: Set this to `True` to treat zero as missing value.
            feature_pre_filter: Set this to `True` to ignroe the features that are unsplittable based
                on `min_data_in_leaf`.
            num_class: Only used in multiclass classification.
            is_unbalance: Used only in `binary` and `multiclassova` classification application, used
                to handle unbalanced training data. Mutually exclusive with `scale_pos_weight`.
            scale_pos_weight: Used only in `binary` and `multiclassova` classification application, specifies the
                ratio of positive class weight to negative class weight. Mutually exclusive with `is_unbalance`.
            sigmoid: Used only in `binary` and `multiclassova` classification and in `lambdarank` application,
                specifies the sigmoid parameter.
            boost_from_average: Used only in `regression`, `binary`, `multiclassova` and `cross-entropy` applications,
                adjusts initial score to the mean of labels for faster convergence.
            reg_sqrt: Used only in `regression` application, specifies whether to use `sqrt(label)` as labels.
            alpha: Used only in `huber` and `quantile` `regression` application, specifies the alpha parameter.
            fair_c: Used only in `fair` `regression` application, specifies the parameter for Fair loss.
            poisson_max_delta_step: Used only in `poisson` `regression` application, specifies the parameter
                for Poisson regression to safeguard optimization.
            tweedie_variance_power: Used only in `tweedie` `regression` application. Used to control the variance of
                the Tweedie distribution. Set closer to 2 to shift towards a Gamma distribution. Set closer to 1 to
                shift towards a Poisson distribution.
            lambdarank_truncation_level: Used only in `lambdarank` application, controls the number of top-results
                to focus on during training.
            lambdarank_norm: Set this to `True` to normalize the lambdas for different queries, and improve the
                performance for unbalanced data. Set this to `False` to enforce the original lambdarank algorithm.
                Used only in `lambdarank` application.
            metric: Metrics to be evaluated on the evaluation set(s).
            metric_freq: The frequency of metric output, every `metric_freq` iterations.
            is_provide_training_metric: Set this to `True` to output the metric result over training data.
            eval_at: Used only with `ndcg` and `map` metrics, specifies the evaluation positions.
            multi_error_top_k: Used only in `multi_error` metric, specifies the threshold for top-k error metric.
            auc_mu_weights: Used only in `auc_mu` metric, specifies the weights for `auc_mu` metric.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.model import LGBMModel
            >>> df = vaex.ml.datasets.load_iris()
            >>> df_train, df_test = df.ml.random_split(test_size=0.20, verbose=False)
            >>> data_schema = DataSchema(
            ...    numerical=["sepal_length", "sepal_width", "petal_length", "petal_width"],
            ... )
            >>> model = LGBMModel(
            ...     cache_directory=".",
            ...     target="class_",
            ...     features=["sepal_width", "petal_length", "petal_width"],
            ...     num_iterations=100,
            ...     learning_rate=0.1,
            ...     num_leaves=31,
            ...     random_state=42,
            ... )
            >>> booster, df_train_pred, df_test_pred = model.fit_transform(data_schema, df_train, df_test, {})
        """
        super().__init__(cache_directory, features, ignore_features, cache_size)
        lgb.register_logger(logger)

        self._verbosity = verbosity
        self._log_evaluation_period = log_evaluation_period
        self._target = target
        self._num_iterations = num_iterations
        self._hyperparameters = {
            "objective": objective,
            "boosting": boosting,
            "data_sample_strategy": data_sample_strategy,
            "learning_rate": learning_rate,
            "num_leaves": num_leaves,
            "tree_learner": tree_learner,
            "random_state": random_state,
            "force_col_wise": force_col_wise,
            "force_row_wise": force_row_wise,
            "histogram_pool_size": histogram_pool_size,
            "max_depth": max_depth,
            "min_data_in_leaf": min_data_in_leaf,
            "min_sum_hessian_in_leaf": min_sum_hessian_in_leaf,
            "bagging_fraction": bagging_fraction,
            "pos_bagging_fraction": pos_bagging_fraction,
            "neg_bagging_fraction": neg_bagging_fraction,
            "bagging_freq": bagging_freq,
            "feature_fraction": feature_fraction,
            "feature_fraction_bynode": feature_fraction_bynode,
            "extra_trees": extra_trees,
            "early_stopping_round": early_stopping_round,
            "first_metric_only": first_metric_only,
            "max_delta_step": max_delta_step,
            "lambda_l1": lambda_l1,
            "lambda_l2": lambda_l2,
            "linear_lambda": linear_lambda,
            "min_gain_to_split": min_gain_to_split,
            "drop_rate": drop_rate,
            "max_drop": max_drop,
            "skip_drop": skip_drop,
            "xgboost_dart_mode": xgboost_dart_mode,
            "uniform_drop": uniform_drop,
            "top_rate": top_rate,
            "other_rate": other_rate,
            "min_data_per_group": min_data_per_group,
            "max_cat_threshold": max_cat_threshold,
            "cat_l2": cat_l2,
            "cat_smooth": cat_smooth,
            "max_cat_to_onehot": max_cat_to_onehot,
            "top_k": top_k,
            "monotone_constraints": monotone_constraints,
            "monotone_constraints_method": monotone_constraints_method,
            "monotone_penalty": monotone_penalty,
            "feature_contri": feature_contri,
            "path_smooth": path_smooth,
            "verbosity": verbosity,
            "use_quantized_grad": use_quantized_grad,
            "num_grad_quant_bins": num_grad_quant_bins,
            "quant_train_renew_leaf": quant_train_renew_leaf,
            "stochastic_rounding": stochastic_rounding,
            "linear_tree": linear_tree,
            "max_bin": max_bin,
            "max_bin_by_feature": max_bin_by_feature,
            "min_data_in_bin": min_data_in_bin,
            "bin_construct_sample_cnt": bin_construct_sample_cnt,
            "is_enable_sparse": is_enable_sparse,
            "enable_bundle": enable_bundle,
            "use_missing": use_missing,
            "zero_as_missing": zero_as_missing,
            "feature_pre_filter": feature_pre_filter,
            "num_class": num_class,
            "is_unbalance": is_unbalance,
            "scale_pos_weight": scale_pos_weight,
            "sigmoid": sigmoid,
            "boost_from_average": boost_from_average,
            "reg_sqrt": reg_sqrt,
            "alpha": alpha,
            "fair_c": fair_c,
            "poisson_max_delta_step": poisson_max_delta_step,
            "tweedie_variance_power": tweedie_variance_power,
            "lambdarank_truncation_level": lambdarank_truncation_level,
            "lambdarank_norm": lambdarank_norm,
            "metric": metric,
            "metric_freq": metric_freq,
            "is_provide_training_metric": is_provide_training_metric,
            "eval_at": eval_at,
            "multi_error_top_k": multi_error_top_k,
            "auc_mu_weights": auc_mu_weights,
        }

    def _fit(
        self,
        data_schema: DataSchema,
        train_dataframe: vaex.DataFrame,
        validation_dataframe: vaex.DataFrame,
        hyperparameters: HyperparametersType | None = None,
    ) -> tuple[lgb.Booster, dict[str, dict[str, list[Any]]]]:
        """Fits the LightGBM model to the given data with the given hyperparameters.

        Args:
            data_schema: The data schema of the dataframes.
            train_dataframe: The training dataframe.
            validation_dataframe: The validation dataframe.
            hyperparameters: The hyperparameters to use for training.

        Raises:
            ValueError: If the target feature is in the feature set.

        Returns:
            The trained LightGBM model.
        """
        if self._target in self._feature_set(data_schema):
            error_msg = f"Target feature {self._target} is in the feature set."
            logger.error(error_msg)
            raise ValueError(error_msg)

        if self._verbosity > 0:
            logger.info("Loading the training dataset into memory.")
        train_dataset = self._load_lgb_dataset(data_schema, train_dataframe)
        if self._verbosity > 0:
            logger.info("Loading the validation dataset into memory.")
        validation_dataset = self._load_lgb_dataset(data_schema, validation_dataframe, reference_dataset=train_dataset)

        if hyperparameters is None:
            hyperparameters = {}
        hyperparameters = {**self._hyperparameters, **hyperparameters}
        self._num_iterations = self._pop_num_iterations(self._hyperparameters)
        if self._verbosity > 0:
            logger.info(
                f"Training the LightGBM model for {self._num_iterations} iterations "
                f"with the following hyperparameters: {hyperparameters}."
            )

        metrics = {}
        callbacks = [lgb.record_evaluation(metrics)]
        if self._verbosity > 0 and self._log_evaluation_period is not None:
            callbacks.append(lgb.log_evaluation(period=self._log_evaluation_period))

        self._model = lgb.train(
            params=hyperparameters,
            train_set=train_dataset,
            num_boost_round=self._num_iterations,
            valid_sets=[train_dataset, validation_dataset],
            valid_names=["train", "validation"],
            callbacks=callbacks,
        )

        return self._model, metrics

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> vaex.DataFrame:
        """Transforms the given dataframe using the LightGBM model.

        Will return the predictions of the model applied to the given dataframe.

        Args:
            data_schema: The data schema of the dataframe.
            dataframe: The dataframe to transform.

        Returns:
            The transformed dataframe.
        """
        if self._verbosity > 0:
            logger.info("Loading the dataset into memory.")
        feature_names = self._feature_set(data_schema)
        dataset = get_columns(dataframe, feature_names).to_pandas_df()

        if self._verbosity > 0:
            logger.info("Transforming the dataset.")
        predictions = self._model.predict(dataset)

        df = dataframe.copy()
        df["prediction"] = predictions
        return df

    def _fingerprint(self) -> Hashable:
        """Returns the fingerprint of the model.

        Appends the target feature and number of iterations to the fingerprint.

        Returns:
            The fingerprint of the model.
        """
        return super()._fingerprint(), self._target, self._num_iterations

    def _default_features(self, data_schema: DataSchema) -> tuple[str, ...]:
        """The default set of features to use for training.

        Args:
            data_schema: The data schema of the dataframes.

        Returns:
            The default set of features.
        """
        features = data_schema.get_features(["numerical", "boolean", "categorical"])
        return tuple(str(feature) for feature in features)

    def _load_lgb_dataset(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame, reference_dataset: lgb.Dataset | None = None
    ):
        """Loads the given dataframe into a LightGBM dataset.

        Args:
            data_schema: The data schema of the dataframe.
            dataframe: The dataframe to load.
            reference_dataset: The reference dataset, required if the dataframe is a validation dataset.

        Returns:
            The LightGBM dataset.
        """
        feature_names = self._feature_set(data_schema)
        df_x = get_columns(dataframe, feature_names).to_pandas_df()
        y = get_column(dataframe, self._target).to_numpy()

        return lgb.Dataset(
            data=df_x,
            label=y,
            reference=reference_dataset,
            categorical_feature=data_schema.get_features(["categorical", "boolean"]),
        )

    def _pop_num_iterations(self, hyperparameters: dict) -> int:
        """Extracts the number of iterations from the hyperparameters.

        Warning:
            All aliases for the number of iterations are removed from the hyperparameters dictionary.
            This is an intentional side effect which is used to avoid passing the number of iterations
            to the LightGBM model twice.

        Args:
            hyperparameters: The hyperparameters dictionary.

        Returns:
            The number of iterations.
        """
        num_iterations_aliases = [
            "num_iterations",
            "num_iteration",
            "n_iter",
            "num_tree",
            "num_trees",
            "num_round",
            "num_rounds",
            "nrounds",
            "num_boost_round",
            "n_estimators",
            "max_iter",
        ]

        valid_aliases = [
            alias
            for alias in num_iterations_aliases
            if alias in hyperparameters and isinstance(hyperparameters[alias], int)
        ]

        num_iterations = hyperparameters[valid_aliases[0]] if valid_aliases else self._num_iterations

        for alias in num_iterations_aliases:
            hyperparameters.pop(alias, None)

        return num_iterations



"""Module for data conversion as a part of the pipeline process.

This module contains the `ConvertStep` class which is a specialized pipeline step for handling data format
conversion. It uses the provided `BaseDataConverter` for converting the data into the desired format.
"""
from __future__ import annotations

from pathlib import Path

from mleko.dataset.convert import BaseConverter
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class ConvertStep(PipelineStep):
    """Pipeline step that manages data conversion from one format to another."""

    _num_inputs = 1
    """Number of inputs expected by the ConvertStep."""

    _num_outputs = 2
    """Number of outputs expected by the ConvertStep."""

    @auto_repr
    def __init__(
        self,
        converter: BaseConverter,
        inputs: list[str] | tuple[str, ...] | tuple[()] = (),
        outputs: list[str] | tuple[str, ...] | tuple[()] = (),
        cache_group: str | None = None,
    ) -> None:
        """Initialize the ConvertStep with the specified data converter.

        Args:
            converter: The DataConverter responsible for handling data format conversion.
            inputs: List or tuple of input keys expected by this step. Should contain a single key,
                corresponding to the list of file Paths to be converted.
            outputs: List or tuple of output keys produced by this step. Should contain two keys,
                corresponding to the DataSchema and DataFrame after conversion.
            cache_group: The cache group to use.
        """
        super().__init__(inputs, outputs, cache_group)
        self._converter = converter

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform data format conversion using the configured converter.

        Args:
            data_container: Contains a list of file Paths to be converted.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            ValueError: If data container contains invalid data - not a list of Paths.

        Returns:
            A DataContainer containing the DataSchema and DataFrame after conversion.
        """
        file_paths = data_container.data[self._inputs[0]]
        if not isinstance(file_paths, list) or not all(isinstance(e, Path) for e in file_paths):
            raise ValueError(f"Invalid data type: {type(file_paths)}. Expected list of Paths.")

        data_schema, dataframe = self._converter.convert(file_paths, self._cache_group, force_recompute)
        data_container.data[self._outputs[0]] = data_schema
        data_container.data[self._outputs[1]] = dataframe
        return data_container
